# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016, 2017, 2018
# --------------------------------------------------------------------------
# Author: Olivier OUDOT, IBM Analytics, France Lab, Sophia-Antipolis

"""
Compiler converting internal model representation to CPO file format.
"""

from docplex.cp.expression import *
from docplex.cp.expression import _domain_min, _domain_max
from docplex.cp.solution import *
from docplex.cp.utils import *
import docplex.cp.config as config
import datetime
import itertools

import sys
import functools


###############################################################################
## Utilities
###############################################################################

# Map of CPO names for each array type
_ARRAY_TYPES = {Type_IntArray: 'intArray', Type_FloatArray: 'floatArray',
                Type_IntExprArray: 'intExprArray', Type_FloatExprArray: 'floatExprArray',
                Type_CumulExprArray: 'cumulExprArray',
                Type_IntervalVarArray: 'intervalVarArray', Type_SequenceVarArray: 'sequenceVarArray',
                Type_IntValueSelectorArray: 'intValueSelectorArray' , Type_IntVarSelectorArray: 'intVarSelectorArray',
                Type_CumulAtomArray: '_cumulAtomArray'}


# Set of CPO types representing an integer
_INTEGER_TYPES = frozenset((Type_Int, Type_PositiveInt, Type_TimeInt))

# Symbol used to reference anonymous variables
_ANONYMOUS = "Anonymous"


###############################################################################
## Public classes
###############################################################################

class CpoCompiler(object):
    """ Compiler to CPO file format """
    __slots__ = ('model',                    # Source model
                 'parameters',               # Solving parameters
                 'add_source_location',      # Indicator to add location traces in generated output
                 'format_version',           # Output format version
                 'is_format_less_than_12_8', # Indicates that format version is less than 12.8
                 'is_format_at_least_12_8',  # Indicates that format version is at least 12.8
                 'is_format_at_least_12_9',  # Indicates that format version is at least 12.9
                 'name_all_constraints',     # Indicator to fore a name on each constraint (for conflict refiner)
                 'min_length_for_alias',     # Minimum variable name length to replace it by an alias
                 'verbose_output',           # Verbose output (not short_output)
                 'id_printable_strings',     # Dictionary of printable string for each identifier
                 'last_location',            # Last source location (file, line)
                 'list_consts',              # List of constants
                 'list_vars',                # List of variables
                 'list_exprs',               # List of model expressions
                 'list_phases',              # List of search phases
                 'expr_infos',               # Map of expression infos.
                                             # Key is expression id, value is list [expr, location, ref_count, cpo_name, is_root, is_compiled]
                 'expr_by_names',            # Map of expressions by CPO name. Used to retrieve expressions from solutions.
                 'alias_name_map',           # Map of variable renamed with a shorter name. key = old name, value = new name
                 )

    def __init__(self, model, **kwargs):
        """ Create a new compiler

        Args:
            model:  Source model
        Optional args:
            context:             Global solving context. If not given, context is the default context that is set in config.py.
            params:              Solving parameters (CpoParameters) that overwrites those in solving context
            add_source_location: Add source location into generated text
            length_for_alias:    Minimum name length to use shorter alias instead
            (others):            All other context parameters that can be changed
        """
        super(CpoCompiler, self).__init__()

        # Build effective context
        if model:
            mparams = model.get_parameters()
            if mparams:
                pparams = kwargs.get('params')
                if pparams:
                    mparams = mparams.clone()
                    mparams.add(pparams)
                kwargs['params'] = mparams
        context = config._get_effective_context(**kwargs)

        # Initialize processing
        self.model = model
        self.parameters = context.params

        self.min_length_for_alias = None
        self.id_printable_strings = {}
        self.name_all_constraints = False
        self.verbose_output = False

        # Set model parameters
        mctx = context.model
        if mctx is not None:
            self.min_length_for_alias = mctx.length_for_alias
            self.name_all_constraints = mctx.name_all_constraints
            self.verbose_output = not mctx.short_output

        # Determine CPO format version
        self.format_version = None if model is None else model.get_format_version()
        if self.format_version is None:
            self.format_version = mctx.version

        # Determine output variant triggers
        if self.format_version is None:
            # Assume most recent solver
            self.is_format_less_than_12_8 = False
            self.is_format_at_least_12_8 = True
            self.is_format_at_least_12_9 = True
        else:
            self.is_format_less_than_12_8 = self.format_version and compare_natural(self.format_version, '12.8') < 0
            self.is_format_at_least_12_8 = self.format_version and compare_natural(self.format_version, '12.8') >= 0
            self.is_format_at_least_12_9 = self.format_version and compare_natural(self.format_version, '12.9') >= 0

        # Initialize source location
        if self.verbose_output:
            if (self.parameters is not None) and (self.parameters.UseFileLocations is not None):
                self.add_source_location = (self.parameters.UseFileLocations in ('On', True))
            elif (mctx is not None) and (mctx.add_source_location is not None):
                self.add_source_location = mctx.add_source_location
            else:
                self.add_source_location = True
        else:
            self.add_source_location = False
        self.last_location = None

        # Initialize expression dictionaries
        self.expr_by_names = {}
        self.expr_infos = {}
        self.list_consts = []
        self.list_vars = []
        self.list_exprs = []
        self.list_phases = []
        self.alias_name_map = {}

        # Precompile model
        self._pre_compile_model()


    def get_expr_map(self):
        """ Get the map of model expressions.

        Returns:
            Map of model expressions.
            Key is the name that has been used to represent the expression in the CPO file format,
            Value is the corresponding model expression.
        """
        return self.expr_by_names


    def print_model(self, out=None):
        """ Compile the model and print the CPO file format in a given output.

        If the given output is a string, it is considered as a file name that is opened by this method
        using 'utf-8' encoding.

        DEPRECATED. Use :meth:`write` instead.

        Args:
            out: Target output, stream or file name. Default is sys.stdout.
        """
        self.write(out)


    def get_as_string(self):
        """ Compile the model in CPO file format into a string

        Returns:
            String containing the model
        """
        # Print the model into a string
        out = StringIO()
        self.write(out)
        res = out.getvalue()
        out.close()

        # Convert in unicode if required
        if IS_PYTHON_2 and (type(res) is str):
            res = unicode(res)
        return res


    def write(self, out=None):
        """ Write the model in CPO format.

        If the given output is a string, it is considered as a file name that is opened by this method
        using 'utf-8' encoding.

        Args:
            out (Optional): Target output stream or file name. If not given, default value is sys.stdout.
        """
        # Check file
        if is_string(out):
            with open_utf8(os.path.abspath(out), mode='w') as f:
                self.write(f)
                return
        # Check default output
        if out is None:
            out = sys.stdout

        # Initialize processing
        model = self.model
        self.last_location = None

        # Write header
        if self.verbose_output:
            banner = u"/" * 79 + "\n"
            mname = model.get_name()
            sfile = model.get_source_file()
            out.write(banner)
            datestr = datetime.datetime.now().strftime("%Y.%m.%d-%H:%M:%S")
            if mname:
                out.write(u"// CPO file generated at {} for model: {}\n".format(datestr, mname))
            else:
                out.write(u"// CPO file generated at {} for anonymous model\n".format(datestr))
            if sfile:
                out.write(u"// Source file: {}\n".format(sfile))
            out.write(banner)

            # Write version if any
            if self.format_version is not None:
                out.write(u"\n//--- Internals ---\n")
                out.write(u"internals {\n")
                out.write(u"   version({});\n".format(self.format_version))
                out.write(u"}\n")

        # Print renamed variables as comment
        snm = self.alias_name_map
        if snm and self.verbose_output:
            out.write(u"\n//--- Aliases ---\n")
            out.write(u"// To reduce CPO file size, the following aliases have been used to replace names longer than {}\n".format(self.min_length_for_alias))
            lvars = sorted(snm.keys(), key=functools.cmp_to_key(lambda v1, v2: compare_natural(v1, v2)))
            for v in lvars:
                out.write(u"// {} = {}\n".format(v, snm[v]))

        # Write constants
        if self.verbose_output:
            out.write(u"\n//--- Constants ---\n")
        for lx in self.list_consts:
            self._write_expression(out, lx)

        # Write variables
        if self.verbose_output:
            out.write(u"\n//--- Variables ---\n")
        for lx in self.list_vars:
            self._write_expression(out, lx)

        # Write expressions
        if self.verbose_output:
            out.write(u"\n//--- Expressions ---\n")
        self.last_location = None
        for lx in self.list_exprs:
            self._write_expression(out, lx)

        # Write KPIs if any
        kpis = model.get_kpis()
        if kpis:
            if self.is_format_at_least_12_9:
                if self.verbose_output:
                    out.write(u"\n//--- KPIs ---\n")
                out.write(u"KPIs {\n")
                for k, (x, l) in kpis.items():
                    # Skip lambda expressions
                    if not isinstance(x, CpoExpr):
                        continue
                    # Write KPI name
                    out.write(self._get_id_string(k));
                    # Retrieve expression infos
                    # if x.name != k:
                    #     out.write(u" = " + self._compile_expression(x, False))
                    # xinfo = self.expr_infos.get(id(x))
                    # if xinfo:
                    #     xn = xinfo[0].name
                    #     if ()
                    #     # Check if expression is already compiled
                    #     if xinfo[5] and xn != k:
                    #         self._write_expression(out, [CpoAlias(x, k), l, 1, self._get_id_string(k), False, False])
                    #     else:
                    #         if xn is not None:
                    #             if xinfo[3] is None:
                    #                xinfo[3] = self._get_id_string(k)
                    #             self._write_expression(out, xinfo)
                    #         else:
                    #             self._write_expression(out, [x, l, 1, self._get_id_string(k), True, False])
                    # else:
                    #     self._write_expression(out, [x, l, 1, self._get_id_string(k), True, False])
                    out.write(u";\n")
                out.write(u"}\n")
            else:
                # Check that all KPI model expressions are integer variables
                for (x, l) in kpis.values():
                    if isinstance(x, CpoExpr) and not isinstance(x, CpoIntVar):
                        raise CpoException("With CPO format version {}, KPI expressions must all be integer variables"
                                           .format(self.format_version))

        # Write search phases
        if self.list_phases:
            if self.verbose_output:
                out.write(u"\n//--- Search phases ---\n")
            out.write(u"search {\n")
            for lx in self.list_phases:
                self._write_expression(out, lx)
            out.write(u"}\n")

        # Write starting point
        spoint = model.get_starting_point()
        if spoint is not None:
            if self.verbose_output:
                out.write(u"\n//--- Starting point ---\n")
            if self.last_location is not None:
                out.write(u"#line off\n")
            out.write(u"startingPoint {\n")
            for var in spoint.get_all_var_solutions():
                self._write_starting_point(out, var)
            out.write(u"}\n")

        # Write parameters
        if self.parameters:
            # Build list of valid parameters
            params = []
            for k in sorted(self.parameters.keys()):
                v = self.parameters[k]
                if v is not None:
                    params.append((k, v))
            if len(params) > 0:
                if self.verbose_output:
                    out.write(u"\n//--- Parameters ---\n")
                if self.last_location is not None:
                    out.write(u"#line off\n")
                out.write(u"parameters {\n")
                for k, v in params:
                    out.write(u"   {} = {};\n".format(k, v))
                out.write(u"}\n")

        # Flush stream (required on Linux rhel6.7)
        out.flush()


    def _write_expression(self, out, xinfo):
        """ Write model expression

        Args:
            out:   Target output
            xinfo: Expression info, list (expr, location, ref_count, cpo_name, is_root, is_compiled)
        """
        # Retrieve expression elements
        expr, loc, rcnt, name, isroot, iscpld = xinfo
        #print("Write expr: {}, refcnt: {}, name: {}, root: {}, iscpld: {}".format(expr, rcnt, name, isroot, iscpld))

        # Trace location if required
        self._write_source_location(out, loc)

        # Write expression
        if iscpld and name:
            out.write(name + u";\n")
        else:
            if name:
                if isinstance(expr, CpoAlias):
                    # Simple alias
                    out.write(name + u" = " + self._compile_expression(expr.expr, False) + u";\n")
                elif isroot and expr.type in (Type_Constraint, Type_SearchPhase, Type_BoolExpr):
                    # Named constraint
                    if self.is_format_at_least_12_8:
                        out.write(name + u": " + self._compile_expression(expr, True) + u";\n")
                    else:
                        out.write(name + u" = " + self._compile_expression(expr, True)+ u";\n" + name + u";\n")
                else:
                    out.write(name + u" = " + self._compile_expression(expr, True)+ u";\n")
            else:
                out.write(self._compile_expression(expr, True) + u";\n")
            # Mark as compiled
            xinfo[5] = True


    def _write_starting_point(self, out, var):
        """ Write a starting point variable

        Args:
            out:  Target output
            var:  Variable solution
        """
        # Build starting point declaration
        cout = []
        if isinstance(var, CpoIntVarSolution):
            self._compile_integer_var_starting_point(var, cout)
        elif isinstance(var, CpoIntervalVarSolution):
            self._compile_interval_var_starting_point(var, cout)
        else:
            #raise CpoException("Internal error: unsupported starting point variable: " + str(var))
            pass

        # Write variable starting point
        if cout:
            out.write(self._get_expr_id(var.expr) + u" = " + u''.join(cout) + u";\n")


    def _write_source_location(self, out, loc):
        """ Write a source location

        Args:
            out:  Target output
            loc:  Source location
        """
        if self.add_source_location:
            lloc = self.last_location
            if loc != lloc:
                if loc is None:
                    out.write( u"#line off\n")
                else:
                    (file, line) = loc
                    lline = u"#line " + str(line)
                    if (lloc is None) or (file != lloc[0]):
                        lline += u' "' + file.replace('\\', '/') + '"'
                    out.write(lline + u"\n")
                self.last_location = loc


    def _get_id_string(self, id):
        """ Get the string representing an identifier

        Args:
            id: Identifier name
        Returns:
            Printable identifier string (including double quotes and escape sequences if needed)
        """
        # Check if already converted
        res = self.id_printable_strings.get(id)
        if res is None:
            # Convert id into string and store result for next call
            res = to_printable_string(id)
            self.id_printable_strings[id] = res
        return res


    def _get_expr_id(self, expr):
        """ Get the printable id of an expression

        Args:
            expr: Expression
        Returns:
            Printable identifier string (including double quotes and escape sequences if needed)
        """
        # Retrieve expression info
        xinfo = self.expr_infos.get(id(expr))
        if xinfo:
            # Expression in a compiled model
            return xinfo[3]
        # Expression out of a model
        xname = expr.get_name()
        if xname:
            return self._get_id_string(xname)
        return _ANONYMOUS


    def _compile_expression(self, expr, root=True):
        """ Compile an expression in a string in CPO format

        Args:
            expr: Expression to compile
            root: Root expression indicator
        Returns:
            String representing this expression in CPO format
        """
        # Initialize working variables
        cout = []  # Result list of strings
        estack = [[expr, 0, False]]  # Expression stack [Expression, child index, parenthesis]

        # Loop while expression stack is not empty
        while estack:
            # Get expression to compile
            edscr = estack[-1]
            e = edscr[0]

            # Check if expression is named and not root (named expression and variable)
            xinfo = self.expr_infos.get(id(e))
            if xinfo:
                ename = xinfo[3]
            else:
                ename = e.name
                if ename:
                    ename = self._get_id_string(ename)
            if ename and (not root or (e is not expr) or isinstance(e, CpoAlias)):
                cout.append(ename)
                estack.pop()
                continue

            # Check constant expressions
            t = e.type
            if t.is_constant:
                estack.pop()
                if t.is_array:
                    vals = e.value
                    if len(vals) == 0:
                        cout.append(_ARRAY_TYPES[t])
                        cout.append("[]")
                    else:
                        cout.append('[')
                        self._compile_integer_var_domain(vals, cout)
                        cout.append(']')
                elif t is Type_Bool:
                    self._compile_boolean_constant(e, cout)
                elif t is Type_TransitionMatrix:
                    self._compile_transition_matrix(e, cout)
                elif t is Type_TupleSet:
                    self._compile_tuple_set(e, cout)
                elif t is Type_StepFunction:
                    self._compile_step_function(e, cout)
                elif t is Type_SegmentedFunction:
                    self._compile_segmented_function(e, cout)
                else:
                    cout.append(_number_value_string(e.value))

            # Check variables
            elif t.is_variable:
                estack.pop()
                if t is Type_IntVar:
                    self._compile_integer_var(e, cout)
                elif t is Type_IntervalVar:
                    self._compile_interval_var(e, cout)
                elif t is Type_SequenceVar:
                    self._compile_sequence_var(e, cout)
                elif t is Type_StateFunction:
                    self._compile_state_function(e, cout)
                elif t is Type_FloatVar:
                    self._compile_float_var(e, cout)

            # Check array
            elif t.is_array:
                oprnds = e.children
                alen = len(oprnds)
                if alen == 0:
                    cout.append(_ARRAY_TYPES[t])
                    cout.append("[]")
                    estack.pop()
                else:
                    cnx = edscr[1]
                    if cnx <= 0:
                        cout.append("[")
                    if cnx >= alen:
                        cout.append("]")
                        estack.pop()
                    else:
                        edscr[1] += 1
                        if cnx > 0:
                            cout.append(", ")
                        estack.append([oprnds[cnx], 0, False])

            # General expression
            else:
                # Get operation elements
                oper = e.operation
                prio = oper.priority
                oprnds = e.children
                oplen = len(oprnds)
                cnx = edscr[1]

                # Check if function call
                if prio < 0:
                    # Check first call
                    if cnx <= 0:
                        # Check special case of sum of cumulexpr
                        if self.is_format_less_than_12_8 and oper is Oper_sum and e.type is Type_CumulExpr:
                            # Replace function call by a serie of additions
                            largs = oprnds[0].value
                            res = largs[0]
                            for v in largs[1:]:
                                res = CpoFunctionCall(Oper_plus, Type_CumulExpr, (res, v))
                            estack[-1] = [res, 0, False]
                            continue

                        cout.append(oper.keyword)
                        cout.append("(")
                    if cnx >= oplen:
                        cout.append(")")
                        estack.pop()
                    else:
                        edscr[1] += 1
                        if cnx > 0:
                            cout.append(", ")
                        estack.append([oprnds[cnx], 0, False])

                # Write operation
                else:
                    # Check parenthesis required
                    parents = edscr[2]

                    # Write operation
                    if cnx <= 0:
                        if oplen == 1:
                            cout.append(oper.keyword)
                        if parents:
                            cout.append("(")
                    if cnx >= oplen:
                        # All operands have been processed
                        if parents:
                            cout.append(")")
                        estack.pop()
                    else:
                        # Process operand
                        edscr[1] += 1
                        if cnx > 0:
                            # Add operator
                            cout.append(" " + oper.keyword + " ")
                        # Check if operand will require to have parenthesis
                        arg = oprnds[cnx]
                        nprio = arg.priority
                        # Parenthesis required if priority is greater than parent node, or if this node is not first child
                        chparnts = (nprio > prio) \
                                  or (nprio >= 5) \
                                  or ((nprio == prio) and (cnx > 0)) \
                                  or ((oplen == 1) and not parents and oprnds[0].children)
                        # Put operand on stack
                        estack.append([arg, 0, chparnts])

        # Check output exists
        if not cout:
            # Raise exception without calling str on expr (recursion)
            raise CpoException("Internal error: unable to compile expression of type {}".format(type(expr)))
        return u''.join(cout)


    def _compile_boolean_constant(self, v, cout):
        """ Compile a boolean constant in a string in CPO format
        Args:
            v:    Constant value
            cout: Output string list
        """
        cout.append("true()" if v.value else "false()")


    def _compile_integer_var(self, v, cout):
        """ Compile a integer variable in a string in CPO format
        Args:
            v:    Variable
            cout: Output string list
        """
        cout.append("intVar(")
        self._compile_integer_var_domain(v.get_domain(), cout)
        cout.append(")")


    def _compile_float_var(self, v, cout):
        """ Compile a float variable in a string in CPO format
        Args:
            v:    Variable
            cout: Output string list
        """
        cout.append("floatVar(")
        cout.append(_number_value_string(v.get_domain_min()))
        cout.append(", ")
        cout.append(_number_value_string(v.get_domain_max()))
        cout.append(")")


    def _compile_interval_var(self, v, cout):
        """ Compile a IntervalVar in a string in CPO format
        Args:
            v:    Variable
            cout: Output string list
        """
        cout.append("intervalVar(")
        args = []
        if v.is_absent():
            args.append("absent")
        elif v.is_optional():
            args.append("optional")
        if v.start != DEFAULT_INTERVAL:
            args.append("start=" + _interval_var_domain_string(v.start))
        if v.end != DEFAULT_INTERVAL:
            args.append("end=" + _interval_var_domain_string(v.end))
        if v.length != DEFAULT_INTERVAL:
            args.append("length=" + _interval_var_domain_string(v.length))
        if v.size != DEFAULT_INTERVAL:
            args.append("size=" + _interval_var_domain_string(v.size))
        if v.intensity is not None:
            args.append("intensity=" + self._compile_expression(v.intensity, root=False))
        if v.granularity is not None:
            args.append("granularity=" + str(v.granularity))
        cout.append(", ".join(args) + ")")


    def _compile_integer_var_starting_point(self, v, cout):
        """ Compile a integer variable starting point in a string in CPO format
        Args:
            v:    Variable solution (CpoIntVarSolution)
            cout: Output string list
        """
        cout.append("(")
        dom = v.value
        dmin = _domain_min(dom)
        dmax = _domain_max(dom)
        cout.append(_int_var_value_string(dmin))
        if dmin != dmax:
            cout.append('..')
            cout.append(_int_var_value_string(dmax))
        cout.append(")")


    def _compile_interval_var_starting_point(self, v, cout):
        """ Compile a starting IntervalVar in a string in CPO format
        Args:
            v:    Variable solution (CpoIntervalVarSolution)
            cout: Output string list
        """
        if v.is_absent():
            cout.append("absent")
            return
        cout.append("(")
        cout.append("present" if v.is_present() else "optional")
        rng = v.get_start()
        if rng is not None:
            cout.append(", start=" + _interval_var_domain_string(rng))
        rng = v.get_end()
        if rng is not None:
            cout.append(", end=" + _interval_var_domain_string(rng))
        rng = v.get_size()
        if rng is not None:
            cout.append(", size=" + _interval_var_domain_string(rng))
        cout.append(")")


    def _compile_sequence_var(self, sv, cout):
        """ Compile a SequenceVar in a string in CPO format
        Args:
            sv:   Sequence variable
            cout: Output string list
        """
        cout.append("sequenceVar(")
        lvars = sv.get_interval_variables()
        if len(lvars) == 0:
            cout.append("intervalVarArray[]")
        else:
            cout.append("[" + ", ".join(self._get_expr_id(v) for v in lvars) + "]")
        types = sv.get_types()
        if types is not None:
            if len(lvars) == 0:
                cout.append(", intArray[]")
            else:
                cout.append(", [" + ", ".join(str(t) for t in types) + "]")
        cout.append(")")


    def _compile_state_function(self, stfct, cout):
        """ Compile a State in a string in CPO format

        Args:
           stfct: Segmented function
           cout:  Output string list
        """
        cout.append("stateFunction(")
        trmx = stfct.get_transition_matrix()
        if trmx is not None:
            cout.append(self._compile_expression(trmx, root=False))
        cout.append(")")


    @staticmethod
    def _compile_transition_matrix(tm, cout):
        """ Compile a TransitionMatrix in a string in CPO format

        Args:
            tm:   Transition matrix
            cout: Output string list
        """
        cout.append("transitionMatrix(")
        cout.append(", ".join(str(v) for v in tm.get_all_values()))
        cout.append(")")


    def _compile_tuple_set(self, e, cout):
        """ Compile a TupleSet in a string in CPO format

        Args:
           e:    Tuple set expression
           cout: Output string list
        """
        tplset = e.value
        if tplset:
            cout.append("[")
            for i, tpl in enumerate(tplset):
                if i > 0:
                    cout.append(", ")
                cout.append("[")
                self._compile_integer_var_domain(tpl, cout)
                cout.append("]")
            cout.append("]")
        else:
            cout.append("tupleSet[]")


    @staticmethod
    def _compile_integer_var_domain(dom, cout):
        """ Compile a integer variable domain in CPO format

        Args:
            dom:   Variable domain
            cout:  Output string list
        """
        if is_array(dom):
            for i, d in enumerate(dom):
                if i > 0:
                    cout.append(", ")
                if isinstance(d, (list, tuple)):
                    cout.append(_int_var_value_string(d[0]) + ".." + _int_var_value_string(d[1]))
                else:
                    cout.append(_number_value_string(d))
        else:
            cout.append(_number_value_string(dom))


    @staticmethod
    def _compile_list_of_integers(lint, cout):
        """ Compile a list of integers in CPO format

        Args:
            lint:  List of integers
            cout:  Output string list
        """
        llen = len(lint)
        i = 0
        while i < llen:
            if i > 0:
                cout.append(", ")
            j = i + 1
            while (j < llen) and (lint[j] == lint[j - 1] + 1):
                j += 1
            if j > i + 1:
                cout.append(str(lint[i]) + ".." + str(lint[j - 1]))
            else:
                cout.append(str(lint[i]))
            i = j


    @staticmethod
    def _compile_step_function(stfct, cout):
        """ Compile a StepFunction in a string in CPO format

        Args:
           stfct: Step function
           cout:  Output string list
        """
        cout.append("stepFunction(")
        for i, s in enumerate(stfct.get_step_list()):
            if i > 0:
                cout.append(", ")
            cout.append('(' + _number_value_string(s[0]) + ", " + str(s[1]) + ')')
        cout.append(")")


    @staticmethod
    def _compile_segmented_function(sgfct, cout):
        """ Compile a SegmentedFunction in a string in CPO format

        Args:
           sgfct: Segmented function
           cout:  Output string list
        """
        cout.append("segmentedFunction(")
        cout.append(", ".join(map(to_string, sgfct.get_segment_list())))
        cout.append(")")


    def _pre_compile_model(self):
        """ Pre-compile model

        This method pre-compile model, compute dependencies between expressions and allocate names if required.

        Result is set in compiler attributes.
        """
        # Check null model
        model = self.model
        if model is None:
            return

        # Scan all expressions
        self.expr_infos = expr_infos = {}
        all_exprs = []
        self._scan_expressions(model.get_all_expressions(), expr_infos, all_exprs)

        # Create an alias list for KPIs
        kpiexprs = []
        kpis = model.get_kpis()
        if kpis:
            for k, (x, l) in kpis.items():
                if isinstance(x, CpoExpr) and not isinstance(x, CpoAlias):
                    if x.get_name() != k:
                        kpiexprs.append((CpoAlias(x, k), l))
                    elif not isinstance(x, CpoVariable):
                        xnfo = expr_infos.get(id(x))
                        if (xnfo is None) or (xnfo[2] == 0):
                            kpiexprs.append((x, l))

        # Add KPIs expressions
        self._scan_expressions(kpiexprs, expr_infos, all_exprs)
        # kpexprs = []
        # self._scan_expressions([(x, l) for (x, l) in kpis.values() if isinstance(x, CpoExpr)], expr_infos, kpexprs)
        # for xnfo in kpexprs:
        #     x = xnfo[0]
        #     if x.name and ((not id(x) in expr_infos) or (x.type.is_variable and xnfo[2] == 1)):
        #         all_exprs.append(xnfo)

        # Initialize lists of expressions
        list_consts = []
        list_vars = []
        vars_set = set()
        list_exprs = []
        list_phases = []
        alias_name_map = {}       # List of variable with a name but renamed shortly
        name_constraints = self.name_all_constraints
        min_length_for_alias = self.min_length_for_alias

        # Create name allocators
        id_allocators = [IdAllocator('_EXP_')] * len(ALL_TYPES)
        id_allocators[Type_IntVar.id] = IdAllocator('_INT_')
        id_allocators[Type_IntervalVar.id] = IdAllocator('_ITV_')
        id_allocators[Type_SequenceVar.id] = IdAllocator('_SEQ_')
        id_allocators[Type_StateFunction.id] = IdAllocator('_FUN_')
        id_allocators[Type_Constraint.id] = IdAllocator('_CTR_')

        # Initialize map of names
        expr_by_names = self.expr_by_names = {}

        # Force to keep original name for KPIs
        for x, l in kpiexprs:
            expr_by_names[x.get_name()] = x

        # print ("All expressions infos to process:")
        # for xinfo in all_exprs:
        #     print("   {}".format(xinfo))
        #
        # Allocate names and split variables and constants
        for xinfo in all_exprs:
            expr = xinfo[0]
            typ = expr.type
            # Allocate name if needed
            xname = expr.get_name()
            if xname:
                if xinfo[5]:
                    # Already compiled, retrieve stored name for the same expression
                    xinfo[3] = expr_infos[id(expr)][3]
                else:
                    # Check if name too long
                    if min_length_for_alias is not None and len(xname) >= min_length_for_alias and not xname in kpis:
                        xname = _allocate_expr_id(id_allocators[typ.id], expr_by_names)
                        alias_name_map[expr.get_name()] = xname
                        xinfo[3] = xname
                    else:
                        # Check if already used elsewhere
                        # if xname in expr_by_names:
                        ox = expr_by_names.get(xname)
                        if not (ox is None or ox is expr):  # To process the case of pre-allocated names of KPIs
                            # Allocate a next instance
                            ndx = 1
                            xname += "@"
                            nname = xname + "1"
                            while nname in expr_by_names:
                                ndx += 1
                                nname = xname + str(ndx)
                            xname = nname
                        xinfo[3] = to_printable_string(xname)
                    expr_by_names[xname] = expr
            elif (xinfo[3] is None) and ((xinfo[2] > 1) or typ.is_variable):
                # Allocate name
                xname = _allocate_expr_id(id_allocators[typ.id], expr_by_names)
                expr_by_names[xname] = expr
                xinfo[3] = xname
            elif name_constraints and xinfo[4] and (typ in (Type_Constraint, Type_BoolExpr)):
                xname = _allocate_expr_id(id_allocators[typ.id], expr_by_names)
                expr_by_names[xname] = expr
                xinfo[3] = xname

            # Split in different expression lists
            if typ.is_constant and xname:
                list_consts.append(xinfo)
            elif typ.is_variable and not isinstance(expr, CpoAlias):
                if expr not in vars_set:
                    vars_set.add(expr)
                    list_vars.append(xinfo)
            elif typ == Type_SearchPhase:
                list_phases.append(xinfo)
            elif xname or xinfo[4] or xinfo[2] > 1:
                list_exprs.append(xinfo)

        # Sort list of constants and variables by natural name order
        sortkey = functools.cmp_to_key(lambda v1, v2: compare_expressions(v1[0], v2[0]))
        self.list_consts = sorted(list_consts, key=sortkey)
        self.list_vars = sorted(list_vars, key=sortkey)

        # Set other attributes
        self.list_exprs = list_exprs
        self.list_phases = list_phases
        self.alias_name_map = alias_name_map


    def _scan_expressions(self, lexpr, expr_infos, all_exprs):
        """ Scan a list of expressions
        Args:
            lexpr:       List of expression to scan (list of tuples (expression, location))
            expr_infos:  Map if expression infos to update
            all_exprs:   List of all root expressions to update
        """
        # Scan all expressions
        for expr, loc in lexpr:
            eid = id(expr)
            xinfo = expr_infos.get(eid)
            if xinfo is not None:
                # Expression already compiled, add it again
                xinfo[2] += 1  # Increment reference count
            else:
                # Create new expression info
                xinfo = [expr, loc, 1, None, True, False]
                expr_infos[eid] = xinfo
                # Process children
                if expr.children:
                    # Initialize expression stack with [expression, child index]
                    stack = [[x, 0] for x in expr.children]
                    while stack:
                        xdscr = stack[-1]
                        e, cx = xdscr
                        # Skip constants
                        if e.type.is_constant_atom and not e.name:
                            stack.pop()
                            continue
                        # Check if expression already processed
                        xid = id(e)
                        xnfo = expr_infos.get(xid)
                        if xnfo:
                            xnfo[2] += 1
                            stack.pop()
                            continue
                        # Check if all children processed
                        if cx >= len(e.children):
                            # Add current children node and remove it from stack
                            xnfo = [e, loc, 1, None, False, False]
                            expr_infos[xid] = xnfo
                            all_exprs.append(xnfo)
                            stack.pop()
                        else:
                            # Push current child on the stack
                            stack.append([e.children[cx], 0])
                            xdscr[1] += 1
            # Append to the list of expressions
            all_exprs.append(xinfo)


###############################################################################
## Public functions
###############################################################################

def get_cpo_model(model, **kwargs):
    """ Convert a model into a string with CPO file format.

    Args:
        model:  Source model
    Optional args:
        context:             Global solving context. If not given, context is the default context that is set in config.py.
        params:              Solving parameters (CpoParameters) that overwrites those in solving context
        add_source_location: Add source location into generated text
        length_for_alias:    Minimum name length to use shorter alias instead
        (others):            All other context parameters that can be changed
    Returns:
        String of the model in CPO file format
    """
    cplr = CpoCompiler(model, **kwargs)
    return cplr.get_as_string()


def expr_to_string(expr):
    """ Convert a single expression into a string using CPO format

    Args:
        expr:  Expression to convert into string
    Returns:
        String of the expression in CPO file format
    """
    return CpoCompiler(None)._compile_expression(expr)


###############################################################################
## Private functions
###############################################################################

_NUMBER_CONSTANTS = {INT_MIN: "intmin", INT_MAX: "intmax",
                     INTERVAL_MIN: "intervalmin", INTERVAL_MAX: "intervalmax",
                     POSITIVE_INFINITY: "inf", NEGATIVE_INFINITY: "-inf", True: "1", False: "0"}

def _number_value_string(val):
    """ Build the string representing a number value

    This methods checks for special values INT_MIN, INT_MAX, INTERVAL_MIN and INTERVAL_MAX.

    Args:
        val: Integer value
    Returns:
        String representation of the value
    """
    try:
        s = _NUMBER_CONSTANTS.get(val)
        return s if s else str(val)
    except:
        # Case where value is not hashable, like numpy.ndarray that can be the value type
        # when numpy operand appears in the left of an overloaded operator.
        return str(val)


def _int_var_value_string(ibv):
    """ Build the string representing an integer variable domain value

    This methods checks for special values INT_MIN and INT_MAX.

    Args:
        ibv: Integer value value
    Returns:
        String representation of the value
    """
    if ibv == INT_MIN:
        return "intmin"
    elif ibv == INT_MAX:
        return "intmax"
    else:
        return str(ibv)


def _allocate_expr_id(allocator, exprmap):
    """ Allocate a new expression id checking it is not already used.

    Args:
        allocator:  Id allocator
        exprmap:    Map of existing expression names
    Returns:
        New id not in exprmap
    """
    id = allocator.allocate()
    while id in exprmap:
        id = allocator.allocate()
    return id


def _interval_var_value_string(ibv):
    """ Build the string representing an interval variable domain value

    This methods checks for special values INTERVAL_MIN and INTERVAL_MAX.

    Args:
        ibv: Interval value
    Returns:
        String representation of the value
    """
    if ibv == INTERVAL_MIN:
        return "intervalmin"
    elif ibv == INTERVAL_MAX:
        return "intervalmax"
    else:
        return str(ibv)


def _interval_var_domain_string(dom):
    """ Build the string representing an interval_var domain

    Args:
        dom: Domain interval
    Returns:
        String representation of the domain
    """
    if isinstance(dom, (list, tuple)):
        smn, smx = dom
        if smn == smx:
            return _interval_var_value_string(smn)
        return _interval_var_value_string(smn) + ".." + _interval_var_value_string(smx)
    else:
        return _interval_var_value_string(dom)





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2018
# --------------------------------------------------------------------------
# gendoc: ignore

'''Runtime Environment
'''
import os
import sys
import traceback

from docplex.util.environment import get_environment, WorkerEnvironment, \
    Environment, LocalEnvironment, OverrideEnvironment


from .project_handler import ProjectZipHandler, ProjectDirectoryHandler


def get_line_of_model(n, handler, model, scenario):
    with handler.get_input_stream(model, scenario, 'model.py') as m:
        lines = m.readlines()
        return lines[n - 1]


class InterpreterError(Exception):
    pass


class _WorkerEnvironmentOverride(WorkerEnvironment):
    '''A worker environment override, only useful when the default_env
    is a LocalEnvironment. If the default_env is a WorkerEnvironment,
    just use it.
    '''
    def __init__(self, env, solve_hook=None):
        super(_WorkerEnvironmentOverride, self).__init__(None)
        self.env = env

    def get_available_core_count(self):
        return self.env.get_available_core_count()

    def get_input_stream(self, name):
        return self.env.get_input_stream(name)

    def get_output_stream(self, name):
        return self.env.get_output_stream(name)

    def set_output_attachment(self, name, filename):
        return self.env.set_output_attachment(name, filename)

    def get_parameter(self, name):
        return self.env.get_parameter(name)

    def update_solve_details(self, details):
        return self.env.update_solve_details(details)

    def notify_start_solve(self, solve_details):
        return self.env.notify_start_solve(solve_details)

    def notify_end_solve(self, status):
        return self.env.notify_end_solve(status)

    def set_stop_callback(self, cb):
        return self.env.set_stop_callback(cb)

    def get_stop_callback(self):
        return self.env.get_stop_callback()


class Executor(object):
    '''The class to execute some code with an execution context that mimics
    that on DODS.

    Note that for a model to be actually run, pandas is necessary.
    '''
    def __init__(self, path):
        self.path = path
        handler = None
        if ProjectZipHandler.accepts(path):
            handler = ProjectZipHandler(path)
        elif ProjectDirectoryHandler.accepts(path):
            handler = ProjectDirectoryHandler(path)
        else:
            raise ValueError('No handler for %s' % path)
        self.handler = handler

    def run_model(self, model, scenario, is_dods=True):
        '''Run the scenario from the specified model.

        Returns:
            a dict with the output tables.
        '''
        variables = {}
        variables['inputs'] = self.handler.get_inputs(model, scenario)
        variables['output_lock'] = get_environment().output_lock
        variables['outputs'] = {}

        override_env = None
        if isinstance(get_environment(), LocalEnvironment):
            override_env = _WorkerEnvironmentOverride(get_environment())

        with OverrideEnvironment(override_env):
            saved_environ = os.environ.copy()
            try:
                if is_dods:
                    os.environ['IS_DODS'] = 'True'

                with self.handler.get_input_stream(model, scenario, 'model.py') as m:
                    try:
                        exec(m.read().decode('utf-8'), variables)
                    except SyntaxError as err:
                        error_class = err.__class__.__name__
                        detail = err.args[0]
                        line_number = err.lineno
                        imsg = 'File "model.py", line %s\n' % line_number
                        imsg += err.text.rstrip() + '\n'
                        spaces = ' ' * (err.offset - 1) if err.offset > 1 else ''
                        imsg += spaces + "^\n"
                        imsg += '%s: %s\n' % (error_class, detail)
                        raise InterpreterError(imsg)
                    except Exception as err:
                        error_class = err.__class__.__name__
                        detail = err.args[0]
                        _, _, tb = sys.exc_info()
                        ttb = traceback.extract_tb(tb)
                        ttb[1] = ('model.py', ttb[1][1], ttb[1][2],
                                  get_line_of_model(ttb[1][1],
                                                    self.handler, model, scenario))
                        line_number = ttb[1][1]
                        ttb = ttb[1:]
                        s = traceback.format_list(ttb)
                        imsg = (''.join(s))
                        imsg += '%s: %s\n' % (error_class, detail)
                        raise InterpreterError(imsg)
                return variables['outputs']
            finally:
                os.environ = saved_environ



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2018
# --------------------------------------------------------------------------

# gendoc: ignore
from docplex.mp.params.parameter_hierarchy_12620 import make_root_params_12620
from docplex.mp.params.parameter_hierarchy_12630 import make_root_params_12630
from docplex.mp.params.parameter_hierarchy_12700 import make_root_params_12700
from docplex.mp.params.parameter_hierarchy_12710 import make_root_params_12710
from docplex.mp.params.parameter_hierarchy_12800 import make_root_params_12800
from docplex.mp.params.parameter_hierarchy_12900 import make_root_params_12900
from docplex.mp.params.parameter_hierarchy_121000 import make_root_params_121000
from docplex.mp.params.parameter_hierarchy_20100 import make_root_params_20100


def _make_default_parameters():
    return make_root_params_121000()


def get_params_from_cplex_version(cpx_version):
    # INTERNAL
    # returns a parameter tree depending on the cplex version, if any.
    # if none is found, returns a default version.
    if cpx_version is None:
        # this can happen, protect from startswith failure
        return _make_default_parameters()
    if cpx_version.startswith("12.6.2."):
        return make_root_params_12620()
    elif cpx_version.startswith("12.6.3."):
        return make_root_params_12630()
    elif cpx_version.startswith("12.7.0"):
        return make_root_params_12700()
    elif cpx_version.startswith("12.7.1"):
        return make_root_params_12710()
    elif cpx_version.startswith("12.8.0"):
        return make_root_params_12800()
    elif cpx_version.startswith("12.9.0"):
        return make_root_params_12900()
    elif cpx_version.startswith("12.10.0"):
        return make_root_params_121000()
    elif cpx_version.startswith("20.1.0"):
        return make_root_params_20100()
    else:
        return _make_default_parameters()



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2019
# --------------------------------------------------------------------------
'''Configuration of Mathematical Programming engine.

The :class:`~Context` is the base class to control the behaviour of solve
engine.

It is not advised to instanciate a Context in your code.

Instead, you should obtain a Context by the following ways:

    * create one using Context.make_default_context()
    * use the one in your :class:`docplex.mp.model.Model`

``docplex.mp`` configuration files are stored in files named:

    * cplex_config.py
    * cplex_config_<hostname>.py
    * docloud_config.py

When obtaining a Context with make_default_context(), the PYTHONPATH is
searched for the configuration files and read.

Configuration files are evaluated with a `context` object in their
scope, and you set values from this context::

    context.solver.docloud.url = 'https://your.application.url.ibm.com'
    context.solver.docloud.key = 'This is an api_key'
    context.cplex_parameters.emphasis.memory = 1
    context.cplex_parameters.emphasis.mip = 2
'''
import os

from copy import deepcopy
import six
from six import iteritems
import shlex
import socket
import sys
import warnings
from os.path import isfile, isabs

from docplex.util.environment import get_environment

from docplex.mp.utils import is_string, open_universal_newline
from docplex.mp.environment import Environment
from docplex.mp.params.cplex_params import get_params_from_cplex_version
from docplex.mp.params.parameters import RootParameterGroup
from docplex.mp.utils import DOcplexException
from docplex.mp.error_handler import docplex_fatal

try:
    from docplex.worker.solvehook import get_solve_hook
except ImportError:
    get_solve_hook = None


def init_cplex_parameters(x):
    return x.init_cplex_parameters()

class StreamWithCustomClose(object):
    # wrapper for streams, so that we can keep track of those who need
    # to be closed by us.
    def __init__(self, target):
        self._target = target
        
    def __getattr__(self, name):
        return getattr(self._target, name)

    def custom_close(self):
        return self._target.close()

    def __str__(self):
        return "<{0}>".format(self._target.name or "input")


# some utility methods
# def _get_value_as_int(d, option):
#     try:
#         value = int(d[option])
#     except Exception:
#         value = None
#     return value
#
#
# def _convert_to_int(value):
#     if str(value).lower() == 'none':
#         return None
#     try:
#         value = int(value)
#     except Exception:
#         value = None
#     return value
#
#
# def _get_value_as_string(d, option):
#     return d.get(option, None)
#
#
# def _get_value_as_boolean(d, option):
#     try:
#         value = _convert_to_bool(d[option])
#     except Exception:
#         value = None
#     return value


_boolean_map = {'1': True, 'yes': True, 'true': True, 'on': True,
                   '0': False, 'no': False, 'false': False, 'off': False}


def _convert_to_bool(value):
    if value is None:
        return None
    elif is_string(value):
        svalue = str(value).lower()
        if svalue == "none":
            return None
        else:
            bvalue = _boolean_map.get(svalue)
            if bvalue is not None:
                return bvalue
            else:
                raise ValueError('Not a boolean: {0}'.format(value))
    else:
        raise ValueError('Not a boolean: {0}'.format(value))


class InvalidSettingsFileError(Exception):
    '''The error raised when an error occured when reading a settings file.

    *New in version 2.8*
    '''
    def __init__(self, mesg, filename=None, source=None, *args, **kwargs):
        super(InvalidSettingsFileError, self).__init__(mesg)
        self.filename = filename
        self.source = source


def is_auto_publishing_solve_details(context):
    try:
        auto_publish_details = context.solver.auto_publish.solve_details
    except AttributeError:
        try:
            auto_publish_details = context.solver.auto_publish
        except AttributeError:
            auto_publish_details = False
    return auto_publish_details




def check_credentials(context):
    #     Checks if the context has syntactically valid credentials. The context
    #     has valid credentials when it has an `url` and a `key` fields and that
    #     both fields are string.
    #
    #     If the credentials are not defined, `message` contains a message describing
    #     the cause.
    #
    #     Returns:
    #         (has_credentials, message): has_credentials` - True if the context contains syntactical credentials.
    #            and `message`  - contains a message if applicable.
    has_credentials = True
    message = None
    if not context.url or not context.key:
        has_credentials = False
    elif not is_string(context.url):
        message = "DOcplexcloud: URL is not a string: {0!s}".format(context.url)
        has_credentials = False
    elif not is_string(context.key):
        message = "API key is not a string: {0!s}".format(context.key)
        has_credentials = False
    if context.key and has_credentials:
        has_credentials = isinstance(context.key, six.string_types)
    return has_credentials, message


def has_credentials(context):
    # Checks if the context has valid credentials.
    #
    # Returns:
    #    True if the context has valid credentials.
    # ignore message
    has_credentials_, _ = check_credentials(context)
    return has_credentials_


def print_context(context):
    # prints the context.
    def print_r(node, prefix):
        for n in sorted(node):
            if not n.startswith('_'):
                path = ".".join([prefix, n] if prefix else [n])
                if isinstance(node.get(n), (dict, SolverContext)):
                    print("%s  # type: %s" % (path, type(node.get(n)).__name__))
                    print_r(node.get(n), path)
                else:
                    print("%s = %s # type: %s" % (path, node.get(n), type(node.get(n)).__name__))

    print_r(context, "context")


class BaseContext(dict):
    # Class for handling the list of parameters.

    def __init__(self, **kwargs):
        """ Create a new context.

        Args:
            List of ``key=value`` to initialize context with.
        """
        super(BaseContext, self).__init__()
        for k, v in kwargs.items():
            self.set_attribute(k, v)

    def __setattr__(self, name, value):
        self.set_attribute(name, value)

    def __getattr__(self, name):
        return self.get_attribute(name)

    def set_attribute(self, name, value):
        self[name] = value

    def get_attribute(self, name, default=None):
        if name.startswith('__'):
            raise AttributeError
        res = self.get(name, default)
        return res

    def display(self):
        # prints the context.
        def print_r(node, prefix):
            for n in sorted(node):
                if not n.startswith('_'):
                    path = ".".join([prefix, n] if prefix else [n])
                    if isinstance(node.get(n), (dict, SolverContext)):
                        print("%s  # type: %s" % (path, type(node.get(n)).__name__))
                        print_r(node.get(n), path)
                    else:
                        print("%s = %s # type: %s" % (path, node.get(n), type(node.get(n)).__name__))

        print_r(self, "context")


class SolverContext(BaseContext):
    # for internal use

    def __init__(self, **kwargs):
        super(SolverContext, self).__init__(**kwargs)
        self.log_output = False
        self.max_threads = get_environment().get_available_core_count()
        self.auto_publish = create_default_auto_publish_context()
        self.kpi_reporting = BaseContext()
        from docplex.mp.progress import ProgressClock
        self.kpi_reporting.filter_level = ProgressClock.Gap

    def __deepcopy__(self, memo):
        # We override deepcopy here just to make sure that we don't deepcopy
        # file descriptors...
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in iteritems(self):
            # do not duplicate those (io like objects)
            if k == "log_output" and hasattr(v, "write"):
                value = v
            else:
                value = deepcopy(v, memo)
            setattr(result, k, value)
        return result
    
  
    
    def get_log_output_as_stream(self):
        try:
            log_output = self.log_output
        except AttributeError:  # pragma: no cover
            return None

        output_stream = None
        # if log_output is an object with a lower attribute, let's use it
        # as string.lower and check for some known string values
        if hasattr(log_output, "lower"):
            k = log_output.lower()
            if k in _boolean_map:
                if _convert_to_bool(k):
                    output_stream = sys.stdout
                else:
                    output_stream = None
            elif k in ["stdout", "sys.stdout"]:
                output_stream = sys.stdout
            elif k in ["stderr", "sys.stderr"]:
                output_stream = sys.stderr
            else:
                output_stream = open(log_output, 'w')
                output_stream = StreamWithCustomClose(output_stream)
        # if log_output is == to True, just use stdout
        if log_output is True:
            output_stream = sys.stdout
        # if it has a write() attribute, just return it
        elif hasattr(log_output, "write"):
            output_stream = log_output

        return output_stream

    log_output_as_stream = property(get_log_output_as_stream)


class Context(BaseContext):
    """ The context used to control the behavior of solve engine.

    Attributes:
        cplex_parameters: A
           :class:`docplex.mp.params.parameters.RootParameterGroup` to store
           CPLEX parameters.
        solver.auto_publish: If ``True``, a model being solved will automatically
            publish all publishable items (``solve_details``,
            ``result_output``, ``kpis_output``).
        solver.auto_publish.solve_details: if ``True``, solve details are
            published automatically.
        solver.auto_publish.result_output: if not None, the filename where
            solution is saved. This can be a list of filenames if multiple
            solutions are to be published. If True, ``solution.json`` is used.
        solver.auto_publish.kpis_output: if not None, the filename where KPIs
            are saved as a table with KPI name and values. Currently only
            csv files are supported. This can be a list of filenames if
            multiple KPIs files are to be published.
        context.solver.auto_publish.kpis_output_field_name: Name of field for KPI names in KPI output
            table. Defaults to 'Name'
        context.solver.auto_publish.kpis_output_field_value: Name of field for KPI values for KPI output
            table. Defaults to 'Value'
        solver.log_output: This attribute can have the following values:

            * True: When True, logs are printed to sys.out.
            * False: When False, logs are not printed.
            * A file-type object: Logs are printed to that file-type object.

        solver.kpi_reporting.filter_level: Specify the filtering level for kpi reporting.
            If None, no filtering is done. Can take values of
            docplex.mp.progress.KpiFilterLevel or a string representation of
            one of the values of this enum (Unfiltered, FilterObjectiveAndBound,
            FilterObjective)
        solver.docloud: The parent node for attributes controlling the solve on Decision Optimization on Cloud.
        solver.docloud.url: The DOcplexcloud service URL.
        solver.docloud.key: The DOcplexcloud service API key.
        solver.docloud.run_deterministic: Specific engine parameters are uploaded to keep the
            run deterministic.
        solver.docloud.verbose: Makes the connector verbose.
        solver.docloud.timeout: The timeout for requests.
        solver.docloud.waittime: The wait time to wait for jobs to finish.
        solver.docloud.verify: If True, verifies SSL certificates.
        solver.docloud.log_requests: If True, the REST requests are logged.
        solver.docloud.log_poll_interval: The interval for log polling.
        solver.docloud.progress_poll_interval: The interval for progress polling.
        solver.docloud.exchange_format: The exchange format to use.
            When setting the format, you can use the following strings: "lp".
            When getting the format, the property type is
            `docplex.mp.format.ExchangeFormat`.
        solver.docloud.job_parameters: dict of job parameters passed to DOCplexcloud.
    """

    def __init__(self, **kwargs):
        # store env used for initialization
        self['_env_at_init'] = kwargs.get('_env')
        # map lazy members to f(model) (actually f(self) ) returning the
        # initial value
        self['_lazy_members'] = \
            {'cplex_parameters': init_cplex_parameters}
        # initialize fields
        if 'solver_context' in kwargs:
            solver_context = kwargs.pop('solver_context')
        else:
            solver_context = SolverContext(docloud=create_default_docloud_context())

        super(Context, self).__init__(solver=solver_context, docplex_tests=BaseContext())
        # update will also ensure compatibility with older kwargs like
        # 'url' and 'api_key'
        self.update(kwargs, create_missing_nodes=True)

        self.model_build_hook = None

    def init_cplex_parameters(self):
        local_env = self.get('_env_at_init') or Environment.get_default_env()
        cplex_version = local_env.cplex_version
        cplex_parameters = get_params_from_cplex_version(cplex_version)
        return cplex_parameters

    def __getattr__(self, name):
        if name not in self:
            lazy_members = self.get('_lazy_members')
            if lazy_members and name in lazy_members:
                evaluated = lazy_members[name](self)
                self[name] = evaluated
        return self.get_attribute(name)

    def _get_raw_cplex_parameters(self):
        # NON LAZY: may return None
        return self.get('cplex_parameters')

    @staticmethod
    def make_default_context(file_list=None, logger=None, **kwargs):
        """Creates a default context.

        If `file_list` is a string, then it is considered to be the name
        of a config file to be read.

        If `file_list` is a list, it is considered to be a list of names
        of a config files to be read.

        if `file_list` is None or not specified, the following files are
        read if they exist:

            * the PYTHONPATH is searched for the following files:

                * cplex_config.py
                * cplex_config_<hostname>.py
                * docloud_config.py

        Args:
            file_list: The list of config files to read.
            kwargs: context parameters to override. See :func:`docplex.mp.context.Context.update`
        """
        context = Context(**kwargs)
        context.read_settings(file_list=file_list, logger=logger)
        if 'DOCPLEX_CONTEXT' in os.environ:
            values_pairs = []
            for v in shlex.split(os.environ['DOCPLEX_CONTEXT']):
                s = v.split('=', 1)  # max 1 split
                # convert values to bool if relevant
                value = _boolean_map.get(s[1].strip().lower(), s[1])
                values_pairs.append((s[0], value))
                if logger:
                    logger.info('Setting context value %s to %s (from DOCPLEX_CONTEXT env)' % (s[0], value))
            context.update_from_list(values_pairs, logger=logger)
        return context

    def copy(self):
        # Makes a deep copy of the context.
        #
        # Returns:
        #   A deep copy of the context.
        return deepcopy(self)

    def clone(self):
        # Makes a deep copy of the context.
        #
        # Returns:
        #   A deep copy of the context.
        return deepcopy(self)

    def override(self):
        return ContextOverride(self)

    def update_from_list(self, values, logger=None):
        # For each pair of `(name, value)` in values, try to set the
        # attribute.
        for name, value in values:
            try:
                self._set_value(self, name, value)
            except AttributeError:
                if logger is not None:
                    logger.warning("Ignoring undefined attribute : {0}".format(name))

    def _set_value(self, root, property_spec, property_value):
        property_list = property_spec.split('.')
        property_chain = property_list[:-1]
        to_be_set = property_list[-1]
        o = root
        for c in property_chain:
            o = getattr(o, c)
        try:
            target_attribute = getattr(o, to_be_set)
        except AttributeError:
            target_attribute = None
        if target_attribute is None:
            # Simply set the attribute
            try:
                setattr(o, to_be_set, property_value)
            except DOcplexException:
                pass  # ignore this
        else:
            # try a set_converted_value if it's a Parameter
            try:
                target_attribute.set(property_value)
            except (AttributeError, TypeError):
                # no set(), just setattr
                setattr(o, to_be_set, property_value)

    def update(self, kwargs, create_missing_nodes=False):
        """ Updates this context from child parameters specified in ``kwargs``.

        The following keys are recognized:

            - cplex_parameters: A set of CPLEX parameters to use instead of the parameters defined as ``context.cplex_parameters``.
            - agent: Changes the ``context.solver.agent`` parameter.
                Supported agents include:

                - ``docloud``: forces the solve operation to use DOcplexcloud
                - ``local``: forces the solve operation to use native CPLEX

            - url: Overwrites the URL of the DOcplexcloud service defined by ``context.solver.docloud.url``.
            - key: Overwrites the authentication key of the DOcplexcloud service defined by ``context.solver.docloud.key``.
            - log_output: if ``True``, solver logs are output to stdout.
                If this is a stream, solver logs are output to that stream object.
                Overwrites the ``context.solver.log_output`` parameter.

        Args:
            kwargs: A ``dict`` containing keyword args to use to update this context.
            create_missing_nodes: When a keyword arg specify a parameter that is not already member of this context,
                creates the parameter if ``create_missing_nodes`` is True.

        """
        for k in kwargs:
            value = kwargs.get(k)
            if value is not None:
                self.update_key_value(k, value,
                                      create_missing_nodes=create_missing_nodes)

    cplex_parameters_key = "cplex_parameters"

    def update_cplex_parameters(self, arg_params):
        # INTERNAL
        if isinstance(arg_params, RootParameterGroup):
            self.cplex_parameters = arg_params
        else:
            new_params = self.cplex_parameters.copy()
            # try a dictionary of parameter qualified names, parameter values
            # e.g. cplex_parameters={'mip.tolerances.mipgap': 0.01, 'timelimit': 180}
            try:
                for pk, pv in iteritems(arg_params):
                    p = new_params.find_parameter(key=pk)
                    if not p:
                        docplex_fatal('Cannot find matching parameter from: {0!r}'.format(pk))
                    else:
                        p.set(pv)
                self.cplex_parameters = new_params

            except (TypeError, AttributeError):
                docplex_fatal('Expecting CPLEX parameters or dict, got: {0!r}'.format(arg_params))

    def update_key_value(self, k, value, create_missing_nodes=False, warn=True):
        if k == 'docloud_context':
            warnings.warn('docloud_context is deprecated, use context.solver.docloud instead')
            self.solver.docloud = value
        elif k == 'cplex_parameters':
            if isinstance(value, RootParameterGroup):
                self.cplex_parameters = value
            else:
                self.update_cplex_parameters(value)

        elif k == 'url':
            self.solver.docloud.url = value
        elif k == 'api_key' or k == 'key':
            self.solver.docloud.key = value
        elif k == 'log_output':
            self.solver.log_output = value
        elif k  == 'override':
            self.update_from_list(iteritems(value))
        elif k == 'proxies':
            self.solver.docloud.proxies = value
        elif k ==  '_env':
            # do nothing this is just here to avoid creating too many envs
            pass
        elif k ==  'agent':
            self.solver.agent = value
        else:
            if create_missing_nodes:
                self[k] = value
            elif warn:
                warnings.warn("Unknown quick-setting in Context: {0:s}, value: {1!s}".format(k, value),
                              stacklevel=2)

    def read_settings(self, file_list=None, logger=None):
        """Reads settings for a list of files.

        If `file_list` is a string, then it is considered to be the name
        of a config file to be read.

        If `file_list` is a list, it is considered to be a list of names
        of config files to be read.

        if `file_list` is None or not specified, the following files are
        read if they exist:

            * the PYTHONPATH is searched for the following files:

                * cplex_config.py
                * cplex_config_<hostname>.py
                * docloud_config.py

        Args:
            file_list: The list of config files to read.

        Raises:
            InvalidSettingsFileError: If an error occurs while reading a config
                file. *(Since version 2.8)*
        """
        if file_list is None:
            file_list = []
            targets = ['cplex_config.py',
                       'cplex_config_{0}.py'.format(socket.gethostname()),
                       'docloud_config.py'
                       ]
            for target in targets:
                if isabs(target) and isfile(target) and target not in file_list:
                    file_list.append(target)
                else:
                    for d in sys.path:
                        f = os.path.join(d, target)
                        if os.path.isfile(f):
                            abs_name = os.path.abspath(f)
                            if abs_name not in file_list:
                                file_list.append(f)

            if len(file_list) == 0:
                file_list = None  # let read_settings use its default behavior

        if isinstance(file_list, six.string_types):
            file_list = [file_list]

        if file_list is not None:
            for f in file_list:
                if os.path.isfile(f):
                    if logger:
                        logger.info("Reading settings from %s" % f)
                    if f.endswith(".py"):
                        self.read_from_python_file(f)

    def read_from_python_file(self, filename):
        # Evaluates the content of a Python file containing code to set up a
        # context.
        #
        # Args:
        #    filename (str): The name of the file to evaluate.
        try:
            if os.path.isfile(filename):
                with open_universal_newline(filename, 'r') as f:
                    l = {'context': self,
                         '__file__': os.path.abspath(filename)}
                    exec(f.read(), globals(), l)
        except Exception as exc:
            raise InvalidSettingsFileError('Error occured while reading file: %s' % filename,
                                           filename=filename, source=exc)
        return self


def create_default_auto_publish_context(defaults=True):
    auto_publish = BaseContext()
    # in a future version, we might want to be able to set individual
    # default values with a dict => that's why we compare to True and False
    if defaults is True and get_solve_hook:
        # Set default values when in a worker
        auto_publish.solve_details = True
        auto_publish.result_output = 'solution.json'
        auto_publish.kpis_output = 'kpis.csv'
        auto_publish.kpis_output_field_name = 'Name'
        auto_publish.kpis_output_field_value = 'Value'
        auto_publish.relaxations_output = 'relaxations.csv'
        auto_publish.conflicts_output = 'conflicts.csv'
    else:
        auto_publish.solve_details = False
        auto_publish.result_output = None
        auto_publish.kpis_output = None
        auto_publish.kpis_output_field_name = None
        auto_publish.kpis_output_field_value = None
        auto_publish.relaxations_output = None
        auto_publish.conflicts_output = None
    return auto_publish

def create_default_docloud_context():
    # for internal use.
    # Returns a context to use as the context.solver.docloud member.
    #
    # This is a Context with predefined fields.
    #
    dctx = BaseContext()
    # There'se a bunch of properties defined so that we can set
    # any values
    dctx.url = None
    dctx.key = None
    dctx.run_deterministic = False
    dctx.verbose = False
    dctx.timeout = None
    dctx.waittime = None
    dctx.verify = None  # default is None so that we use defaults
    dctx.log_requests = None
    dctx.exchange_format = None
    dctx.debug_dump = None
    dctx.debug_dump_dir = None
    dctx.log_poll_interval = None
    dctx.progress_poll_interval = None
    dctx.verbose_progress_logger = None
    dctx.delete_job = True
    # if true, download job info after solve() has finished and fire
    # the last details as a progress_info. Mostly for debug.
    dctx.fire_last_progress = False
    # Mostly for debug: This callback is called when the solve is finished.
    # It should be a method taking **kwargs. It will be called with those kwargs:
    # - jobid: the jobid
    # - client: the docloud client used to connect to docloud
    # - connector: the DOcloudConnector
    dctx.on_solve_finished_cb = None
    # The proxies
    dctx.proxies = None
    # additional job parameters
    dctx.job_parameters = None
    # mangle names into x<...?>
    dctx.mangle_names = False
    return dctx


class ContextOverride(Context):

    def __init__(self, initial_context):
        soc2 = deepcopy(initial_context.solver)
        super(ContextOverride, self).__init__(solver_context=soc2)
        self._initial_context = initial_context  # unchanged
        self.cplex_parameters = initial_context.cplex_parameters

    # @property
    # def solver(self):
    #     if self._solver is None:
    #         self._solver = deepcopy(self._initial_context.solver)
    #     return self._solver

    def update_key_value(self, k, value, create_missing_nodes=False, warn=True):
        if k == 'docloud_context':
            warnings.warn('docloud_context is deprecated, use context.solver.docloud instead')
            self.solver.docloud = value
        elif k == 'cplex_parameters':
            if isinstance(value, RootParameterGroup):
                self.cplex_parameters = value
            else:
                self.update_cplex_parameters(value)
        elif k == 'time_limit':
            time_limit_failed = True
            try:
                time_limit = int(value)
                if time_limit >= 0:
                    # this method makes a local copy for the duration of solve()
                    self.update_cplex_parameters({"timelimit": time_limit})
                    time_limit_failed = False

            except ValueError:
                pass
            if time_limit_failed:
                print("Invalid time limit: {0!r} - ignored".format(value))

        elif k == 'url':
            self.solver.docloud.url = value
        elif k == 'api_key' or k == 'key':
            self.solver.docloud.key = value
        elif k == 'log_output':
            self.solver.log_output = value
        elif k == 'override':
            self.update_from_list(iteritems(value))
        elif k == 'proxies':
            self.solver.docloud.proxies = value
        elif k == '_env':
            # do nothing this is just here to avoid creating too many envs
            pass
        elif k == 'agent':
            self.solver.agent = value
        else:
            if create_missing_nodes:
                self.k = value
            elif warn:
                warnings.warn("Unknown quick-setting in Context: {0:s}, value: {1!s}".format(k, value),
                              stacklevel=2)

    def update_cplex_parameters(self, arg_params):
        # INTERNAL
        new_params = self.cplex_parameters.copy()
        # try a dictionary of parameter qualified names, parameter values
        # e.g. cplex_parameters={'mip.tolerances.mipgap': 0.01, 'timelimit': 180}
        try:
            for pk, pv in iteritems(arg_params):
                p = new_params.find_parameter(key=pk)
                if not p:
                    docplex_fatal('Cannot find matching parameter from: {0!r}'.format(pk))
                else:
                    p.set(pv)
            self.cplex_parameters = new_params

        except (TypeError, AttributeError):
            docplex_fatal('Expecting CPLEX parameters or dict, got: {0!r}'.format(arg_params))

    def update_from_list(self, key_value_pairs, logger=None):
        # For each pair of `(name, value)` in values, try to set the
        # attribute.
        for name, value in key_value_pairs:
            try:
                self._set_value(self, name, value)
            except AttributeError:
                if logger is not None:
                    logger.warning("Ignoring undefined attribute : {0}".format(name))

    def _set_value(self, root, property_spec, property_value):
        property_list = property_spec.split('.')
        property_chain = property_list[:-1]
        to_be_set = property_list[-1]
        o = root
        for c in property_chain:
            o = getattr(o, c)
        try:
            target_attribute = getattr(o, to_be_set)
        except AttributeError:
            target_attribute = None
        if target_attribute is None:
            # Simply set the attribute
            try:
                setattr(o, to_be_set, property_value)
            except DOcplexException:
                print('attribute not found: {0}'.format(to_be_set))
        else:
            # try a set_converted_value if it's a Parameter
            try:
                target_attribute.set(property_value)
            except (AttributeError, TypeError):
                # no set(), just setattr
                setattr(o, to_be_set, property_value)


# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2018
# --------------------------------------------------------------------------

# gendoc: ignore

'''Since version 2.7, docplex can be used even if docloud is not available.

When docloud is not installed, cloud solving is disabled. Additionnaly,
we redefine our now JobSolveStatus class in this case::

  class JobSolveStatus(Enum):
        """ Job solve status values.

        This `Enum` is used to convert job solve status string values into an
        enumeration::

            >>> job = client.get_job(jobid)
            >>> solveStatus = JobSolveStatus[job['solveStatus']]

        Attributes:
            UNKNOWN: The algorithm has no information about the solution.
            FEASIBLE_SOLUTION: The algorithm found a feasible solution.
            OPTIMAL_SOLUTION: The algorithm found an optimal solution.
            INFEASIBLE_SOLUTION: The algorithm proved that the model is infeasible.
            UNBOUNDED_SOLUTION: The algorithm proved the model unbounded.
            INFEASIBLE_OR_UNBOUNDED_SOLUTION: The model is infeasible or unbounded.
        """
        UNKNOWN = 0
        FEASIBLE_SOLUTION = 1
        OPTIMAL_SOLUTION = 2
        INFEASIBLE_SOLUTION = 3
        UNBOUNDED_SOLUTION = 4
        INFEASIBLE_OR_UNBOUNDED_SOLUTION = 5


If docloud is available, use the `JobSolveStatus <https://api-swagger-oaas.docloud.ibmcloud.com/api_swagger/pythondoc/index.html?cm_mc_uid=82840450017915192306798&cm_mc_sid_50200000=77001241530691867305#docloud.status.JobSolveStatus>`_ from docloud.
'''
try:
    from docloud.status import JobSolveStatus
except ImportError:
    from enum import Enum

    class JobSolveStatus(Enum):
        """ Job solve status values.

        This `Enum` is used to convert job solve status string values into an
        enumeration::

            >>> job = client.get_job(jobid)
            >>> solveStatus = JobSolveStatus[job['solveStatus']]

        Attributes:
            UNKNOWN: The algorithm has no information about the solution.
            FEASIBLE_SOLUTION: The algorithm found a feasible solution.
            OPTIMAL_SOLUTION: The algorithm found an optimal solution.
            INFEASIBLE_SOLUTION: The algorithm proved that the model is infeasible.
            UNBOUNDED_SOLUTION: The algorithm proved the model unbounded.
            INFEASIBLE_OR_UNBOUNDED_SOLUTION: The model is infeasible or unbounded.
        """
        UNKNOWN = 0
        FEASIBLE_SOLUTION = 1
        OPTIMAL_SOLUTION = 2
        INFEASIBLE_SOLUTION = 3
        UNBOUNDED_SOLUTION = 4
        INFEASIBLE_OR_UNBOUNDED_SOLUTION = 5



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

from docplex.mp.utils import is_iterable
from docplex.mp.basic import ModelingObjectBase
from docplex.mp.utils import DOcplexException, is_number
from docplex.mp.compat23 import izip
from docplex.mp.sttck import StaticTypeChecker

import copy


class PwlFunction(ModelingObjectBase):
    """
    This class models piecewise linear (PWL) functions. This class is not intended to be instantiated:
    piecewise linear functions are defined by invoking :func:`docplex.mp.model.Model.piecewise`,
    or :func:`docplex.mp.model.Model.piecewise_as_slopes`.

    Piecewise-linear functions are important in many applications.
    They are often specified either:

    * by giving a set of slopes, a set of breakpoints at which the slopes change, and the value of the functions at
      a given point, or
    * by giving an ordered list of (x,y) points that are linearly connected, along with the slope before the first
      point and the slope after the last point.

    Note that a piecewise-linear function may be discontinuous.
    """

    @staticmethod
    def check_number(logger, arg, caller=None):
        StaticTypeChecker.typecheck_num_nan_inf(logger, arg, caller)


    @staticmethod
    def check_list_pair_breaksxy(logger, arg):
        if not is_iterable(arg):
            logger.fatal("argument 'breaksxy' expects iterable, {0!r} was passed".format(arg))
        if isinstance(arg, tuple):
            # Encapsulate tuple argument into a list: this allows defining a PWL with a tuple if there is only
            #  one element in its definition
            arg = [arg]
        if len(arg) == 0:
            logger.fatal("argument 'breaksxy' must be a non-empty list of (x, y) tuples.")
        prev_pair = None
        pprev_pair = None
        for pair in arg:
            if isinstance(pair, tuple):
                if len(pair) != 2:
                    logger.fatal("invalid tuple in 'breaksxy': {0!s}. Each tuple must have 2 items.".format(pair))
                PwlFunction.check_number(logger, pair[0])
                PwlFunction.check_number(logger, pair[1])
            else:
                logger.fatal("invalid item in 'breaksxy': {0!s}. Each item must be a (x, y) tuple.".format(pair))
            if prev_pair is not None:
                if pair[0] < prev_pair[0]:
                    logger.fatal("X coordinate in: {0!s} cannot be smaller than previous break abscisse: {1!s}.".
                                 format(pair, prev_pair))
                if pprev_pair is not None and pair[0] == prev_pair[0] and prev_pair[0] == pprev_pair[0]:
                    logger.fatal(
                        "invalid break: {0!s}. There cannot be more than 2 consecutive breaks with same abscisse.".
                            format(pair))
            pprev_pair = prev_pair
            prev_pair = pair

    @staticmethod
    def check_number_pair(logger, arg):
        if arg is None:
            logger.fatal("argument 'anchor' must be defined")
        if isinstance(arg, tuple):
            if len(arg) != 2:
                logger.fatal("invalid tuple for 'anchor': {0!s}. Anchor argument must have 2 items.".format(arg))
            PwlFunction.check_number(logger, arg[0])
            PwlFunction.check_number(logger, arg[1])
        else:
            logger.fatal("invalid value for 'anchor': {0!s}. Anchor argument must be a (x, y) tuple.".format(arg))

    @staticmethod
    def check_list_pair_slope_breakx(logger, arg, anchor):
        if arg is None:
            logger.fatal("argument 'slopebreaksx' must be defined")
        if not is_iterable(arg):
            logger.fatal("not an iterable: {0!s}".format(arg))
        if len(arg) == 0:
            return
        if isinstance(arg, tuple):
            # Encapsulate tuple argument into a list: this allows defining a PWL with a tuple if there is only
            #  one element in its definition
            arg = [arg]
        prev_pair = None
        pprev_pair = None
        for pair in arg:
            if isinstance(pair, tuple):
                if len(pair) != 2:
                    logger.fatal("invalid tuple in 'slopebreaksx': {0!s}. Each tuple must have 2 items.".
                                 format(pair))
                PwlFunction.check_number(logger, pair[0])
                PwlFunction.check_number(logger, pair[1])
            else:
                logger.fatal("invalid item in 'slopebreaksx': {0!s}. Each item must be a (x, y) tuple.".format(pair))
            if prev_pair is not None:
                if pair[1] < prev_pair[1]:
                    logger.fatal("X coordinate in: {0!s} cannot be smaller than previous break abscisse: {1!s}.".
                                 format(pair, prev_pair))
                if pprev_pair is not None and pair[1] == prev_pair[1] and prev_pair[1] == pprev_pair[1]:
                    logger.fatal(
                        "invalid break: {0!s}. There cannot be more than 2 consecutive breaks with same abscisse.".
                            format(pair))
                if pair[1] == prev_pair[1] and anchor[0] == pair[1]:
                    logger.fatal("anchor {0!s} cannot be defined at discontinuity point: {1!s}".
                                 format(anchor, pair))
            pprev_pair = prev_pair
            prev_pair = pair

    class _PwlAsBreaks:
        """
        When using this class, the piecewise linear function is specified by:
         - Breakpoints defined as a list of coordinate pairs `(x[i], y[i])` defining the segments of the PWL function.
         - Before the first segment of the PWL function there is a half-line; its slope is specified by `preslope`.
         - After the last segment of the the PWL function there is a half-line; its slope is specified by `postslope`.
        Two consecutive breakpoints may have the same x-coordinate; in such cases there is a discontinuity in the
        PWL function.  Three consecutive breakpoints may not have the same x-coordinate.
        """

        def __init__(self, preslope, breaksxy, postslope):
            self._preslope = preslope
            self._breaksxy = self._reformulate_breaksxy(breaksxy)
            self._postslope = postslope

        @property
        def preslope(self):
            return self._preslope

        @property
        def breaksxy(self):
            return self._breaksxy

        @property
        def postslope(self):
            return self._postslope

        def deepcopy(self):
            breaksxy_copy = copy.deepcopy(self.breaksxy)
            return PwlFunction._PwlAsBreaks(self.preslope, breaksxy_copy, self.postslope)

        @staticmethod
        def _reformulate_breaksxy(breaksxy):
            if isinstance(breaksxy, tuple):
                return [] if len(breaksxy) == 0 else [breaksxy]
            return breaksxy

        @staticmethod
        def _remove_useless_intermediate_breaks(preslope, breaksxy, postslope):
            result_breaksxy = []
            current_slope = preslope
            prev_break = None
            for br in breaksxy:
                if prev_break is None:
                    pass
                else:
                    if br[0] == prev_break[0]:
                        # Check discontinuity
                        if br[1] != prev_break[1]:
                            result_breaksxy.append(prev_break)
                            result_breaksxy.append(br)
                            current_slope = None
                    else:
                        slope = (br[1] - prev_break[1]) / (br[0] - prev_break[0])
                        if current_slope is not None and current_slope != slope:
                            # Add prev_break in list
                            result_breaksxy.append(prev_break)
                        current_slope = slope
                prev_break = br
            # Handle last break
            if not result_breaksxy:
                # Set result breaks = first break
                result_breaksxy = breaksxy[0]
            elif current_slope is not None and current_slope != postslope:
                result_breaksxy.append(prev_break)
            return preslope, result_breaksxy, postslope

        def _get_break_at_index(self, index):
            if len(self.breaksxy) <= index:
                return None, None, index
            break_1 = self.breaksxy[index]
            if len(self.breaksxy) > (index + 1):
                break_2 = self.breaksxy[index + 1]
                if break_1[0] == break_2[0]:
                    # Discontinuity
                    return break_1, break_2, index + 1
            return break_1, None, index

        def _get_y_value(self, x_coord, prev_break_index=-1):
            """
            :param x_coord:
            :param prev_break_index: this parameter is mandatory if a breakxy tuple does exist before x_coord. Otherwise
                an exception is raised.
            :return:
            """
            if prev_break_index < 0:
                break_1, break_2, last_ind = self._get_break_at_index(0)
                if break_1[0] < x_coord:
                    raise DOcplexException("Invalid arguments passed to PwlAsBreaks._get_y_value()")
                if break_1[0] == x_coord:
                    y_coord_1 = break_1[1]
                    y_coord_2 = None if break_2 is None else break_2[1]
                    return y_coord_1, y_coord_2, last_ind
                y_coord_1 = break_1[1] - self.preslope * (break_1[0] - x_coord)
                return y_coord_1, None, -1
            break_1, break_2, last_ind = self._get_break_at_index(prev_break_index)
            next_break_1, next_break_2, next_last_ind = self._get_break_at_index(last_ind + 1)
            if next_break_1 is None:
                # x-coord is after last break
                last_break = break_1 if break_2 is None else break_2
                y_coord_1 = last_break[1] + self.postslope * (x_coord - last_break[0])
                return y_coord_1, None, last_ind
            else:
                if x_coord == break_1[0]:
                    # Here, one must have: x_coord > break_1[0]
                    raise DOcplexException("Invalid arguments passed to PwlAsBreaks._get_y_value()")
                if x_coord == next_break_1[0]:
                    y_coord_1 = next_break_1[1]
                    y_coord_2 = None if next_break_2 is None else next_break_2[1]
                    return y_coord_1, y_coord_2, next_last_ind
                y_coord_prev = break_1[1] if break_2 is None else break_2[1]
                y_coord_next = next_break_1[1]
                slope = (y_coord_next - y_coord_prev) / (next_break_1[0] - break_1[0])
                y_coord_1 = y_coord_prev + slope * (x_coord - break_1[0])
                return y_coord_1, None, last_ind

        def evaluate(self, x_val):
            """ Evaluates the breaks-based PWL function at the point whose x-coordinate is `x_val`.

            Args:
                x_val: The x value for which we want to compute the value of the function.

            Returns:
                The value of the PWL function at point `x_val`
                A DOcplexException exception is raised when evaluating at a discontinuity of the PWL function.
            """
            prev_break_index, index = -1, 0
            while index < len(self.breaksxy):
                break_1, break_2, index = self._get_break_at_index(index)
                if break_1 is None:
                    raise DOcplexException("Invalid PWL definition: no break point is defined")
                if break_1[0] < x_val:
                    prev_break_index = index
                else:
                    if break_1[0] == x_val and break_2 is not None:
                        raise DOcplexException("Cannot evaluate PWL at a discontinuity")
                    break
                index += 1
            y_val, _, _ = self._get_y_value(x_val, prev_break_index)
            return y_val

        def _get_all_breaks(self, all_x_coord):
            all_breaks = []
            prev_break_ind = -1
            for x_coord in all_x_coord:
                y_coord_1, y_coord_2, prev_break_ind = self._get_y_value(x_coord, prev_break_ind)
                all_breaks.append((x_coord, y_coord_1) if y_coord_2 is None else
                                  [(x_coord, y_coord_1), (x_coord, y_coord_2)])
            return all_breaks

        def get_nb_intervals(self):
            nb_discontinuities = 0
            prev_br = None
            for br in iter(self.breaksxy):
                if prev_br is not None and prev_br[0] == br[0]:
                    nb_discontinuities += 1
                prev_br = br
            return len(self.breaksxy) - nb_discontinuities - 1

        def __add__(self, arg):
            if isinstance(arg, PwlFunction._PwlAsBreaks):
                all_x_coord = sorted({br[0] for br in self.breaksxy + arg.breaksxy})
                all_breaks_left = self._get_all_breaks(all_x_coord)
                all_breaks_right = arg._get_all_breaks(all_x_coord)
                result_breaksxy = []
                # Both lists have same size, with same x-coord for breaks ==> perform the addition on each break
                for br_l, br_r in izip(all_breaks_left, all_breaks_right):
                    if isinstance(br_l, tuple) and isinstance(br_r, tuple):
                        result_breaksxy.append((br_l[0], br_l[1] + br_r[1]))
                    else:
                        if isinstance(br_l, tuple):
                            # br_r is a list containing 2 tuple pairs
                            result_breaksxy.append((br_l[0], br_l[1] + br_r[0][1]))
                            result_breaksxy.append((br_l[0], br_l[1] + br_r[1][1]))
                        elif isinstance(br_r, tuple):
                            # br_l is a list containing 2 tuple pairs
                            result_breaksxy.append((br_r[0], br_l[0][1] + br_r[1]))
                            result_breaksxy.append((br_r[0], br_l[1][1] + br_r[1]))
                        else:
                            # br_l and br_r are two lists, each containing 2 tuple pairs
                            result_breaksxy.append((br_l[0][0], br_l[0][1] + br_r[0][1]))
                            result_breaksxy.append((br_l[0][0], br_l[1][1] + br_r[1][1]))
                result_preslope = self.preslope + arg.preslope
                result_postslope = self.postslope + arg.postslope
                return PwlFunction._PwlAsBreaks(*self._remove_useless_intermediate_breaks(
                    result_preslope, result_breaksxy, result_postslope))

            elif is_number(arg):
                return PwlFunction._PwlAsBreaks(
                    self.preslope, [(br[0], br[1] + arg) for br in self.breaksxy], self.postslope)

            else:
                raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))

        def __sub__(self, arg):
            if isinstance(arg, PwlFunction._PwlAsBreaks):
                return self + arg * (-1)
            elif is_number(arg):
                return PwlFunction._PwlAsBreaks(
                    self.preslope, [(br[0], br[1] - arg) for br in self.breaksxy], self.postslope)
            else:
                raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))

        def __mul__(self, arg):
            if is_number(arg):
                return PwlFunction._PwlAsBreaks(*self._remove_useless_intermediate_breaks(
                    self.preslope * arg, [(br[0], br[1] * arg) for br in self.breaksxy], self.postslope * arg))
            else:
                raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))

        def translate(self, arg):
            if is_number(arg):
                return PwlFunction._PwlAsBreaks(
                    self.preslope, [(br[0] + arg, br[1]) for br in self.breaksxy], self.postslope)
            else:
                raise DOcplexException("Invalid type for argument: {0!s}.".format(arg))

        def __str__(self):
            return self.to_string()

        def to_string(self):
            return '({0}, {1}, {2})'.format(self.preslope, self.breaksxy, self.postslope)

        def repr_string(self):
            return 'preslope={0},breaksxy={1},postslope={2}'.format(self.preslope, self.breaksxy, self.postslope)

    class _PwlAsSlopes:
        """
        When using this class, the piecewise linear function is specified by:
         - a list of tuple pairs `(slope[i], breakx[i])` of slopes and x-coordinates defining the slope of the piecewise
           function between the previous breakpoint (or minus infinity if there is none) and the breakpoint with
           x-coordinate `breakx[i]`,
         - the slope after the last specified breakpoint, and
         - the coordinates of the 'anchor point'. The purpose of the anchor point is to ground the piecewise-linear
           function specified by the list of slopes and breakpoints.
        Note that:
         - The `breakx[i]` values must be increasing. If two consecutive `breakx` values have the same value, a
           discontinuity is defined and the value associated with the second argument is considered to be a "step".
         - The list of tuple pairs `(slope[i], breakx[i])` may be empty.
         - The default value for the anchor point is the origin (point with coordinates (0, 0)).
         - If the piecewise linear function defines some discontinuities, the anchor must not reside at one of
           these discontinuities, since the function would not be uniquely defined.
        """

        def __init__(self, slopebreaksx, lastslope, anchor=(0, 0)):
            self._slopebreaksx = self._reformulate_slopebreaksx(slopebreaksx)
            self._lastslope = lastslope
            self._anchor = anchor

        @property
        def slopebreaksx(self):
            return self._slopebreaksx

        @property
        def lastslope(self):
            return self._lastslope

        @property
        def anchor(self):
            return self._anchor

        def deepcopy(self):
            slopebreaksx_copy = copy.deepcopy(self.slopebreaksx)
            anchor_copy = copy.deepcopy(self.anchor)
            return PwlFunction._PwlAsSlopes(slopebreaksx_copy, self.lastslope, anchor_copy)

        @staticmethod
        def _reformulate_slopebreaksx(slopebreaksx):
            if isinstance(slopebreaksx, tuple):
                return [] if len(slopebreaksx) == 0 else [slopebreaksx]
            return slopebreaksx

        @staticmethod
        def _compute_breaksxy_after(slope_breaks, anchor):
            breaks_xy = []
            start_x, start_y = anchor[0], anchor[1]
            for (slope, break_x) in slope_breaks:
                delta_x = break_x - start_x
                start_x = break_x
                if delta_x > 0:
                    start_y = start_y + slope * delta_x
                else:
                    # Discontinuity: slope is considered to be a "step"
                    start_y += slope
                breaks_xy.append((start_x, start_y))
            return breaks_xy

        @staticmethod
        def _compute_breaksxy_before(start_slope, slope_breaks, anchor):
            breaks_xy = []
            start_x, start_y = anchor[0], anchor[1]
            last_slope = start_slope
            for (slope, break_x) in slope_breaks:
                delta_x = break_x - start_x
                start_x = break_x
                if delta_x < 0 or anchor[0] == break_x:
                    start_y = start_y + last_slope * delta_x
                else:
                    # Discontinuity: slope is considered to be a "step"
                    start_y -= last_slope
                last_slope = slope
                breaks_xy.append((start_x, start_y))
            return breaks_xy, last_slope

        def convert_to_pwl_as_breaks(self):
            breaks_before = [(s, b) for (s, b) in self.slopebreaksx if b <= self.anchor[0]]
            breaks_after = [(s, b) for (s, b) in self.slopebreaksx if b > self.anchor[0]]
            # Compute y value at each break point
            anchor_slope = breaks_after[0][0] if len(breaks_after) > 0 else self.lastslope
            breaks_before.reverse()
            breaks_xy_before, preslope = self._compute_breaksxy_before(anchor_slope, breaks_before, self.anchor)
            breaks_xy_before.reverse()
            breaks_xy_after = self._compute_breaksxy_after(breaks_after, self.anchor)
            # Now, we can build the PWL as breaks
            breaksxy = breaks_xy_before + breaks_xy_after
            if len(breaksxy) > 0:
                return PwlFunction._PwlAsBreaks(preslope, breaksxy, self.lastslope)
            else:
                # No breakpoint is defined
                return PwlFunction._PwlAsBreaks(self.lastslope, [self.anchor], self.lastslope)

        def _get_safe_xy_anchor(self):
            """
            Return an anchor point that is on or after (if last break corresponds to a discontinuity) the largest
            x-coord corresponding to a break or the anchor.
            :return:
            """
            breaks_after = [(s, b) for (s, b) in self.slopebreaksx if b > self.anchor[0]]
            breaks_xy_after = self._compute_breaksxy_after(breaks_after, self.anchor)
            if len(breaks_xy_after) > 0:
                # Check if last break corresponds to a discontinuity
                if len(breaks_xy_after) > 1 and breaks_xy_after[-2][0] == breaks_xy_after[-1][0]:
                    # Returns point with x-coord = last_x_coord + 1
                    return breaks_xy_after[-1][0] + 1, breaks_xy_after[-1][1] + self.lastslope
                return breaks_xy_after[-1]
            return self.anchor

        @staticmethod
        def _remove_useless_intermediate_slopes(slopebreaksx, lastslope, anchor):
            result_slopebreaksx = []
            prev_sbr = None
            for sbr in slopebreaksx:
                if prev_sbr is not None:
                    if sbr[0] != prev_sbr[0]:
                        result_slopebreaksx.append(prev_sbr)
                prev_sbr = sbr
            if prev_sbr is not None and prev_sbr[0] != lastslope:
                result_slopebreaksx.append(prev_sbr)
            return result_slopebreaksx, lastslope, anchor

        def _get_all_slopebreaks(self, all_x_coord):
            all_slopebreaks = []
            iter_slopebreakx = iter(self.slopebreaksx)
            current_slopebreakx = next(iter_slopebreakx, None)
            for x_coord in all_x_coord:
                if current_slopebreakx is None:
                    all_slopebreaks.append((self.lastslope, x_coord))
                else:
                    while current_slopebreakx is not None and x_coord > current_slopebreakx[1]:
                        prev_slopebreakx = current_slopebreakx
                        current_slopebreakx = next(iter_slopebreakx, None)
                        if current_slopebreakx is not None and current_slopebreakx[1] == prev_slopebreakx[1]:
                            # Case of a discontinuity ==> update last item in result list to a list containing 2 tuples
                            all_slopebreaks[-1] = [(prev_slopebreakx[0], prev_slopebreakx[1]),
                                                   (current_slopebreakx[0], current_slopebreakx[1])]
                    if current_slopebreakx is None:
                        all_slopebreaks.append((self.lastslope, x_coord))
                    else:
                        all_slopebreaks.append((current_slopebreakx[0], x_coord))
            # Handle case where last break is a discontinuity
            prev_slopebreakx = current_slopebreakx
            current_slopebreakx = next(iter_slopebreakx, None)
            if current_slopebreakx is not None:
                # Case of a discontinuity ==> update last item in result list to a list containing 2 tuples
                all_slopebreaks[-1] = [(prev_slopebreakx[0], prev_slopebreakx[1]),
                                       (current_slopebreakx[0], current_slopebreakx[1])]
            return all_slopebreaks

        def __add__(self, arg):
            if isinstance(arg, PwlFunction._PwlAsSlopes):
                all_x_coord = sorted({sbr[1] for sbr in self.slopebreaksx + arg.slopebreaksx})
                all_slopebreaks_left = self._get_all_slopebreaks(all_x_coord)
                all_slopebreaks_right = arg._get_all_slopebreaks(all_x_coord)
                result_slopebreaksxy = []
                # Both lists have same size, with same x-coord for slopebreaks
                #   ==> perform the addition of slopes on each break
                for sbr_l, sbr_r in izip(all_slopebreaks_left, all_slopebreaks_right):
                    if isinstance(sbr_l, tuple) and isinstance(sbr_r, tuple):
                        result_slopebreaksxy.append((sbr_l[0] + sbr_r[0], sbr_l[1]))
                    else:
                        if isinstance(sbr_l, tuple):
                            # sbr_r is a list containing 2 tuple pairs
                            result_slopebreaksxy.append((sbr_l[0] + sbr_r[0][0], sbr_l[1]))
                            result_slopebreaksxy.append((sbr_r[1][0], sbr_l[1]))
                        elif isinstance(sbr_r, tuple):
                            # sbr_l is a list containing 2 tuple pairs
                            result_slopebreaksxy.append((sbr_l[0][0] + sbr_r[0], sbr_r[1]))
                            result_slopebreaksxy.append((sbr_l[1][0], sbr_r[1]))
                        else:
                            # sbr_l and sbr_r are two lists, each containing 2 tuple pairs
                            result_slopebreaksxy.append((sbr_l[0][0] + sbr_r[0][0], sbr_l[0][1]))
                            result_slopebreaksxy.append((sbr_l[1][0] + sbr_r[1][0], sbr_l[0][1]))
                result_lastslope = self.lastslope + arg.lastslope

                if self.anchor[0] == arg.anchor[0]:
                    result_anchor = (self.anchor[0], self.anchor[1] + arg.anchor[1])
                else:
                    # Compute a new anchor based on the last x-coord in the slopebreakx list + anchor point
                    anchor_l = self._get_safe_xy_anchor()
                    anchor_r = arg._get_safe_xy_anchor()
                    delta = anchor_r[0] - anchor_l[0]
                    if anchor_l[0] < anchor_r[0]:
                        result_anchor = (anchor_r[0], anchor_l[1] + anchor_r[1] + delta * self.lastslope)
                    else:
                        result_anchor = (anchor_l[0], anchor_l[1] + anchor_r[1] - delta * arg.lastslope)

                return PwlFunction._PwlAsSlopes(*self._remove_useless_intermediate_slopes(
                    result_slopebreaksxy, result_lastslope, result_anchor))

            elif is_number(arg):
                return PwlFunction._PwlAsSlopes(copy.deepcopy(self.slopebreaksx),
                                                self.lastslope, (self.anchor[0], self.anchor[1] + arg))
            else:
                raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))

        def __sub__(self, arg):
            if isinstance(arg, PwlFunction._PwlAsSlopes):
                return self + arg * (-1)
            elif is_number(arg):
                return PwlFunction._PwlAsSlopes(copy.deepcopy(self.slopebreaksx),
                                                self.lastslope, (self.anchor[0], self.anchor[1] - arg))
            else:
                raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))

        def __mul__(self, arg):
            if is_number(arg):
                return PwlFunction._PwlAsSlopes(*self._remove_useless_intermediate_slopes(
                    [(br[0] * arg, br[1]) for br in self.slopebreaksx],
                    self.lastslope * arg, (self.anchor[0], self.anchor[1] * arg)))
            else:
                raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))

        def translate(self, arg):
            if is_number(arg):
                return PwlFunction._PwlAsSlopes(
                    [(br[0], br[1] + arg) for br in self.slopebreaksx], self.lastslope,
                    (self.anchor[0] + arg, self.anchor[1]))
            else:
                raise DOcplexException("Invalid type for argument: {0!s}.".format(arg))

        def __str__(self):
            return self.to_string()

        def to_string(self):
            return '{' + ''.join(
                repr(slope) + ' -> ' + repr(break_x) + ';' for (slope, break_x) in self._slopebreaksx) + \
                   repr(self._lastslope) + '}(' + repr(self.anchor[0]) + ', ' + repr(self.anchor[1]) + ')'

    # _name_generator = _AutomaticSymbolGenerator(pattern="pwl", offset=1)

    def __init__(self, model, pwl_def, name=None):
        ModelingObjectBase.__init__(self, model, name=name)
        self._pwl_def = pwl_def
        self._pwl_def_as_breaks = None
        self._set_pwl_definition(pwl_def)

    def _set_pwl_definition(self, pwl_def):
        # INTERNAL
        if isinstance(pwl_def, PwlFunction._PwlAsBreaks):
            # Use the same data structure as input for internal representation (do not duplicate)
            self._pwl_def_as_breaks = pwl_def
        elif isinstance(pwl_def, PwlFunction._PwlAsSlopes):
            pwl_def_as_breaks = pwl_def.convert_to_pwl_as_breaks()
            self._set_pwl_as_breaks(pwl_def_as_breaks.preslope, pwl_def_as_breaks.breaksxy,
                                    pwl_def_as_breaks.postslope)
        else:
            self.model._checker.fatal("Invalid definition for Piecewise Linear Function: {0!s}.".format(pwl_def))

    def _set_pwl_as_breaks(self, preslope, breaksxy=None, postslope=None):
        """Internal format to represent a piecewise linear function is based on the Cplex representation"""
        self._pwl_def_as_breaks = self._PwlAsBreaks(preslope, breaksxy, postslope)

    def copy(self, target_model, var_mapping):
        pwl_def_copy = self.pwl_def.deepcopy()
        return target_model._piecewise(pwl_def_copy, self.get_name())

    @property
    def pwl_def(self):
        return self._pwl_def

    @property
    def pwl_def_as_breaks(self):
        return self._pwl_def_as_breaks

    # __call__ builds an expression equal to the piecewise linear value of its argument, based
    # on the definition of the PWL function.
    #
    # Args:
    #     e: Accepts any object that can be transformed into an expression:
    #         decision variables, expressions, or numbers.
    #
    # Returns:
    #     An expression that can be used in arithmetic operators and constraints.
    #
    # Note:
    #     Building the expression generates one auxiliary decision variable.
    def __call__(self, e):
        self.model._checker.typecheck_operand(e, caller="Model.pwl", accept_numbers=True)
        return self.model._add_pwl_expr(self, e)

    def __hash__(self):
        return id(self)

    def __str__(self):
        return self.pwl_def.__str__()

    def __repr__(self):
        return 'docplex.mp.pwl.PwlFunction({0})'.format(self.pwl_def_as_breaks.repr_string())

    def clone(self):
        """ Creates a copy of the PWL function on the same model.

        Returns:
            The copy of the PWL function.
        """
        return self.copy(self.model, None)

    def __add__(self, e):
        return self.plus(e)

    def __radd__(self, e):
        return self.plus(e)

    def __iadd__(self, e):
        self.model.fatal('Cannot modify a PWL function')
        # self.add(e)
        # return self

    def plus(self, e):
        cloned = self.clone()
        return cloned.add(e)

    def add(self, arg):
        """ Adds an expression to self.

        Note:
            This method does not create a new PWL function but modifies the `self` instance.

        Args:
            arg: The expression to be added. Can be a PWL function or a number.

        Returns:
            The modified self.
        """
        if isinstance(arg, PwlFunction):
            if (isinstance(self.pwl_def, PwlFunction._PwlAsBreaks) and
                    isinstance(arg.pwl_def, PwlFunction._PwlAsBreaks)) or \
                    (isinstance(self.pwl_def, PwlFunction._PwlAsSlopes) and
                         isinstance(arg.pwl_def, PwlFunction._PwlAsSlopes)):
                self._pwl_def = self.pwl_def + arg.pwl_def
                self._set_pwl_definition(self._pwl_def)
            else:
                # Use Breaks representation
                self._pwl_def = self.pwl_def_as_breaks + arg.pwl_def_as_breaks
                self._set_pwl_definition(self._pwl_def)
        elif is_number(arg):
            self._pwl_def = self.pwl_def + arg
            self._set_pwl_definition(self._pwl_def)
        else:
            raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))
        return self

    def __sub__(self, e):
        return self.minus(e)

    def __rsub__(self, e):
        return self * (-1) + e

    def __isub__(self, e):
        self.model.fatal('Cannot modify a PWL function')

    def minus(self, e):
        cloned = self.clone()
        return cloned.subtract(e)

    def subtract(self, arg):
        """ Subtracts an expression from this PWL function.

        Note:
            This method does not create a new function but modifies the `self` instance.

        Args:
            arg: The expression to be subtracted. Can be either a PWL function, or a number.

        Returns:
            The modified self.
        """
        if isinstance(arg, PwlFunction):
            if (isinstance(self.pwl_def, PwlFunction._PwlAsBreaks) and
                    isinstance(arg.pwl_def, PwlFunction._PwlAsBreaks)) or \
                    (isinstance(self.pwl_def, PwlFunction._PwlAsSlopes) and
                         isinstance(arg.pwl_def, PwlFunction._PwlAsSlopes)):
                self._pwl_def = self.pwl_def - arg.pwl_def
                self._set_pwl_definition(self._pwl_def)
            else:
                # Use Breaks representation
                self._pwl_def = self.pwl_def_as_breaks - arg.pwl_def_as_breaks
                self._set_pwl_definition(self._pwl_def)
        elif is_number(arg):
            self._pwl_def = self.pwl_def - arg
            self._set_pwl_definition(self._pwl_def)
        else:
            raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))
        return self

    def __mul__(self, e):
        return self.times(e)

    def __rmul__(self, e):
        return self.times(e)

    def __imul__(self, e):
        self.model.fatal('Cannot modify a PWL function')
        # return self.multiply(e)

    def __div__(self, e):
        return self.quotient(e)

    def __truediv__(self, e):
        # for py3
        # INTERNAL
        return self.quotient(e)  # pragma: no cover

    def __itruediv__(self, e):
        # for py3
        # INTERNAL
        self.model.fatal('Cannot modify a PWL function')  # pragma: no cover

    def __idiv__(self, other):
        self.model.fatal('Cannot modify a PWL function')

    def __rtruediv__(self, e):
        # for py3
        self.fatal("PWL function {0!s} cannot be used as denominator of {1!s}", self, e)  # pragma: no cover

    def __rdiv__(self, e):
        self.fatal("PWL function {0!s} cannot be used as denominator of {1!s}", self, e)

    def quotient(self, e):
        cloned = self.clone()
        cloned.divide(e)
        return cloned

    def divide(self, arg):
        """ Divides this PWL function by a number.

        Note:
            This method does not create a new function but modifies the `self` instance.

        Args:
            arg: The number that is used to divide `self`.

        Returns:
            The modified `self`.
        """
        self.model.typecheck_as_denominator(arg, numerator=self)
        inverse = 1.0 / float(arg)
        return self.multiply(inverse)

    def times(self, e):
        cloned = self.clone()
        return cloned.multiply(e)

    def multiply(self, arg):
        """ Multiplies this PWL function by a number.

        Note:
            This method does not create a new function but modifies the `self` instance.

        Args:
            arg: The number that is used to multiply `self`.

        Returns:
            The modified `self`.
        """
        if is_number(arg):
            self._pwl_def = self.pwl_def * arg
            self._set_pwl_definition(self._pwl_def)
        else:
            raise DOcplexException("Invalid type for right hand side operand: {0!s}.".format(arg))
        return self

    def translate(self, arg):
        """ Translate this PWL function by a number.
        This method creates a new PWL function instance for which all breakpoints have been moved
        along the horizontal axis by the amount specified by `arg`.

        Args:
            arg: The number that is used to translate all breakpoints.

        Returns:
            The translated PWL function.
        """
        if is_number(arg):
            return PwlFunction(self.model, self.pwl_def.translate(arg))
        else:
            raise DOcplexException("Invalid type for argument: {0!s}.".format(arg))

    def evaluate(self, x_val):
        """ Evaluates the PWL function at the point whose x-coordinate is `x_val`.

        Args:
            x_val: The x value for which we want to compute the value of the function.

        Returns:
            The value of the PWL function at point `x_val`.
            A DOcplexException exception is raised when evaluating at a discontinuity of the PWL function.
        """
        return self._pwl_def_as_breaks.evaluate(x_val)

    def plot(self, lx=None, rx=None, k=1, **kwargs):  # pragma: no cover
        """
        This method displays the piecewise linear function using the matplotlib package, if found.

        :param lx: The value to show the `preslope` (must be before the first breakpoint x value).
        :param rx: The value to show the `postslope` (must be after the last breakpoint x value).
        :param k: Scaling factor to calculate default values for `rx` and/or `lx` if these arguments are not provided,
                based on mean interval length between the `x` values of breakpoints.
        :param kwargs: additional arguments to be passed to matplotlib plot() function
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise DOcplexException('matplotlib is required for plot()')
        bks = self.pwl_def_as_breaks.breaksxy
        xs = [bk[0] for bk in bks]
        ys = [bk[1] for bk in bks]
        # compute mean delta_x
        first_x = xs[0]
        last_x = xs[-1]
        nb_intervals = self._pwl_def_as_breaks.get_nb_intervals()
        # k times the mean interval length is used for left/right extra points
        kdx_m = k * (last_x - first_x) / float(nb_intervals) if nb_intervals > 0 else 1

        if lx is None:
            lx = first_x - kdx_m
        ly = ys[0] - self.pwl_def_as_breaks.preslope * (first_x - lx)
        xs.insert(0, lx)
        ys.insert(0, ly)

        if rx is None or rx <= last_x:
            rx = last_x + kdx_m
        ry = ys[-1] + self.pwl_def_as_breaks.postslope * (rx - last_x)
        xs.append(rx)
        ys.append(ry)

        if plt:
            plt.plot(xs, ys, **kwargs)
            if self.name:
                plt.title('pwl: {0}'.format(self.name))
            plt.show()



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2017
# --------------------------------------------------------------------------

# gendoc: ignore

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None


def new_empty_dataframe():
    return DataFrame([])


def make_solution(col_values, col_names, keep_zeros):
    if col_values:
        # fetch all values in one call.
        dd = {'value': col_values}
        if col_names is not None:
            assert len(col_names) == len(col_values)
            dd['name'] = col_names
        ret = DataFrame(dd)
        if not keep_zeros:
            ret = ret[ret['value'] != 0]
            ret = ret.reset_index(drop=True)

        return ret
    else:
        return new_empty_dataframe()


# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
from collections import OrderedDict


class FastOrderedDict(dict):
    """ A subclass of dict that keeps key ordering.

    """
    __slots__ = ('_key_seq',)

    def __init__(self, init_val=None):
        """
        Create a new ordered dictionary. Cannot init from a normal dict,
        nor from kwargs, since items order is undefined in those cases.

        """
        # noinspection PyTypeChecker
        dict.__init__(self)
        self._key_seq = []
        if init_val is None:
            pass
        elif isinstance(init_val, FastOrderedDict):
            self._key_seq = init_val.keys()
            dict.update(self, init_val)
        elif isinstance(init_val, dict):
            self._key_seq.extend(init_val.keys())
            for k in init_val:
                dict.__setitem__(self, k, init_val[k])
        else:
            init_list = list(init_val)
            if init_list:
                if isinstance(init_list[0], tuple):
                    # init_val is assumed to be a list of (key, value) tuples
                    self._update_from_key_values(init_val)
                else:
                    keys = init_val
                    self._update_from_key_values([(k, 1) for k in keys])

    def __delitem__(self, key):
        self.remove(key)

    def remove(self, key):
        """ Removes a key from the dict; updates key sequence
        :param key:
        """
        try:
            dict.__delitem__(self, key)
            # find the index without calling == otherwise create constraints...
            key_index = self._index_is(key)
            if key_index >= 0:
                del self._key_seq[key_index]
        except KeyError:
            pass

    def equals(self, other):
        if len(self) != len(other):
            return False
        self_iteritems = self.iteritems()
        other_iteritems = other.iteritems()
        try:
            while True:
                self_k, self_v = next(self_iteritems)
                other_k, other_v = next(other_iteritems)
                if self_k != other_k:
                    return False
                elif self_v != other_v:
                    return False
                else:
                    continue
        except StopIteration:
            pass
        return True

    def __eq__(self, other):
        if isinstance(other, FastOrderedDict):
            return self.equals(other)
        else:
            return False

    def __repr__(self):
        return '%s([%s])' % (self.__class__.__name__, ', '.join(
            ['(%r, %r)' % (key, self[key]) for key in self._key_seq]))

    def __setitem__(self, key, val, dict_setitem=dict.__setitem__):
        if key not in self:
            self._key_seq.append(key)
        dict_setitem(self, key, val)

    __str__ = __repr__

    def __deepcopy__(self, memo):
        """
        To allow deepcopy to work with OrderedDict.
        """
        from copy import deepcopy

        return self.__class__(deepcopy(self.items(), memo))

    def copy(self):
        """ returns a copy of the ordered dict.
        """
        return FastOrderedDict(self)

    def items(self):
        """ Returns a lits of (key, value) tuples in sequence order.

        """
        return [(k, self[k]) for k in self._key_seq]

    def keys(self):
        """ Returns a copy of the ordered keys.
        """
        return self._key_seq[:]

    # noinspection PyMethodOverriding
    def values(self):
        """ Return a list of values in key sequence order.
        """
        return [self[key] for key in self._key_seq]

    def iteritems(self):
        """ Returns an iterator on key, value pairs, properly ordered
        """

        def generate_items(od):
            keyseq = od._key_seq
            for key in keyseq:
                yield (key, self[key])

        return generate_items(self)

    def iterkeys(self):
        """ Returns an iterator on the ordered keys.
        """
        return iter(self._key_seq)

    __iter__ = iterkeys

    def itervalues(self):
        """ Returns an iterator over values, properly ordered
        """

        def generate_values(od):
            iter_keys = od.iterkeys()
            while True:
                yield od[iter_keys.next()]

        return generate_values(self)

    def clear(self):
        """ Clears the dict, and its key sequence
        """
        dict.clear(self)
        self._key_seq = []

    def update(self, from_arg):
        """
        Update from another OrderedDict or sequence of (key, value) pairs

        """
        if isinstance(from_arg, (FastOrderedDict, OrderedDict)):
            for key, val in from_arg.items():
                self[key] = val
        elif isinstance(from_arg, dict):
            # we lose compatibility with other ordered dict types this way
            raise TypeError('undefined order, cannot get items from dict')
        else:
            self._update_from_key_values(from_arg)

    def _update_from_key_values(self, kv_seq):
        try:
            for key, val in kv_seq:
                self[key] = val
        except (TypeError, ValueError):
            for key in kv_seq:
                self[key] = 1

    def _index_is(self, key):
        # returns the index of key in the seqeunce without calling ==, but "is"
        # returns -1 if not found
        idx = 0
        hash_key = hash(key)
        for k in self._key_seq:
            if k is key or hash(k) == hash_key:
                return idx
            else:
                idx += 1
        else:
            return -1

    def reverse(self):
        """
        Reverse the order of the keys.

        """
        self._key_seq.reverse()

    def sort(self, *args, **kwargs):
        """ Applies a sort to the key sequence.

        The underlying dict is not modified.

        """
        self._key_seq.sort(*args, **kwargs)

    def __reduce__(self):
        'Return state information for pickling'
        items = [(k, self[k]) for k in self]
        # inst_dict = vars(self).copy()
        # for k in vars(OrderedDict()):
        #     inst_dict.pop(k, None)
        # if inst_dict:
        #     return (self.__class__, (items,), inst_dict)
        return self.__class__, (items,)

    def update_from_item_value(self, item, value=1,
                               _dict_get=dict.get, dict_set=dict.__setitem__,
                               fastdict_set=__setitem__):
        """
        This differs from standard Counter when a dict instance is required.
        :param item: the key to be updated
        :param value: the associated value
        :return:
        """
        if value:
            old_value = _dict_get(self, item, 0)
            if old_value:
                new_value = old_value + value
                if 0 != new_value:
                    dict_set(self, item, new_value)
                else:
                    del self[item]
            else:
                # we are sure item is not already present, it must be added to the sequence...
                fastdict_set(self, item, value)


def update_dict_from_item_value(ddict, item, value=1, normalize=False):
    if value:
        if item in ddict:
            new_value = ddict.get(item, 0) + value
            if not normalize or new_value:
                ddict[item] = new_value
            else:
                del ddict[item]
        else:
            ddict[item] = value





# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# ----------------------------------

# gendoc: ignore

import sys

from docplex.mp.error_handler import DOcplexException
from docplex.mp.utils import is_iterable, OutputStreamAdapter
from docplex.mp.compat23 import izip_longest

import json


class SolutionJSONEncoder(json.JSONEncoder):
    # INTERNAL

    def __init__(self, **kwargs):
        # pop kwargs I know before super()
        self.keep_zeros = kwargs.pop("keep_zeros", False)
        super(SolutionJSONEncoder, self).__init__(**kwargs)

    def default(self, solution):
        n = {'CPLEXSolution': self.encode_solution(solution)}
        return n

    def encode_solution(self, solution):
        n = {}
        n["version"] = "1.0"
        n["header"] = self.encode_header(solution)

        m = solution.model
        was_solved = m._has_solution()
        n["variables"] = self.encode_variables(solution, was_solved)
        lc = self.encode_linear_constraints(solution, was_solved)
        if len(lc) > 0:
            n["linearConstraints"] = lc
        qc = self.encode_quadratic_constraints(solution)
        if len(qc) > 0:
            n["quadraticConstraints"] = qc
        return n

    @staticmethod
    def encode_header(solution):
        n = {}
        n["problemName"] = solution.problem_name
        if solution.has_objective():
            n["objectiveValue"] = "{}".format(solution.objective_value)
        n["solved_by"] = solution.solved_by
        return n

    @classmethod
    def encode_linear_constraints(cls, solution, was_solved):
        n = []
        model = solution.model
        if was_solved:
            duals = []
            try:
                if not model._solved_as_mip():
                    duals = model.dual_values(model.iter_linear_constraints())
            except DOcplexException as dex:
                # in some cases, cplex may change the problem type at solve time.
                if "not available for integer" not in str(dex):
                    # ignore "not available for integer problems error, raise others.
                    raise
            slacks = model.slack_values(model.iter_linear_constraints())
        else:
            duals = []
            slacks = []
            model.warning("Model has not been solved, no dual and slack values are printed")

        for (ct, d, s) in izip_longest(model.iter_linear_constraints(),
                                       duals, slacks,
                                       fillvalue=None):
            c = {"name": ct.name,
                 "index": ct.index}
            if s:
                c["slack"] = s
            if d:
                c["dual"] = d
            n.append(c)
        return n

    @classmethod
    def encode_quadratic_constraints(cls, solution):
        n = []
        model = solution.model
        duals = []
        # RTC#37375
        # if not model._solves_as_mip():
        #     duals = model.dual_values(model.iter_quadratic_constraints())
        slacks = []
        was_solved = model._has_solution()
        if was_solved:
            slacks = model.slack_values(model.iter_quadratic_constraints())
        for (ct, d, s) in izip_longest(model.iter_quadratic_constraints(),
                                       duals, slacks,
                                       fillvalue=None):
            # basis status is not yet supported
            c = {"name": ct.name,
                 "index": ct.index}
            if s:
                c["slack"] = s
            if d:
                c["dual"] = d
            n.append(c)
        return n

    def encode_variables(self, sol, was_solved):
        model = sol.model
        n = []
        if was_solved:
            try:
                reduced_costs = [] if model._solved_as_mip() else model.reduced_costs(model.iter_variables())
            except DOcplexException as dex:
                if "not available for integer" in str(dex):
                    reduced_costs = []
                else:
                    raise
        else:
            reduced_costs = []
            model.warning("Model has not been solved, no reduced costs are printed")

        keep_zeros = sol._keep_zeros
        if self.keep_zeros:
            keep_zeros = keep_zeros or self.keep_zeros

        for (dvar, rc) in izip_longest(model.iter_variables(),
                                       reduced_costs,
                                       fillvalue=None):
            if not dvar.is_generated():  # 38934
                value = sol._get_var_value(dvar)
                if keep_zeros or value:
                    v = {"index": "{}".format(dvar.index),
                         "name": dvar.name,
                         "value": "{}".format(value)}
                    if rc is not None:
                        v["reducedCost"] = rc
                    n.append(v)
        return n


class SolutionJSONPrinter(object):
    json_extension = ".json"

    @classmethod
    def print_to_stream2(cls, out, solutions, indent=None, **kwargs):
        # solutions can be either a plain solution or a sequence or an iterator
        sol_to_print = list(solutions) if is_iterable(solutions) else [solutions]
        # encode all solutions in dict ready for json output
        encoder = SolutionJSONEncoder(**kwargs)
        solutions_as_dict = [encoder.default(sol) for sol in sol_to_print]
        # use an output stream adapter for py2/py3 and str/unicode compatibility
        osa = OutputStreamAdapter(out)
        if len(sol_to_print) == 1:  # if only one solution, use at root node
            osa.write(json.dumps(solutions_as_dict[0], indent=indent))
        else:  # for multiple solutions, we want a "CPLEXSolutions" root
            osa.write(json.dumps({"CPLEXSolutions": solutions_as_dict}, indent=indent))

    @classmethod
    def print_to_stream(cls, solutions, out, extension=json_extension, indent=None, **kwargs):
        if out is None:
            # prints on standard output
            cls.print_to_stream2(sys.stdout, solutions, indent=indent, **kwargs)
        elif isinstance(out, str):
            # a string is interpreted as a path name
            path = out if out.endswith(extension) else out + extension
            with open(path, "w") as of:
                cls.print_to_stream(solutions, of, indent=indent, **kwargs)
                # print("* file: %s overwritten" % path)
        else:
            cls.print_to_stream2(out, solutions, indent=indent, **kwargs)




# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2016
# --------------------------------------------------------------------------

# gendoc: ignore
"""
This is the MP package.
"""

import platform
import warnings
import sys

from sys import version_info

ERROR_STRING = "docplex is not compatible with this version of Python: only 64 bits on Windows, Linux, Darwin and AIX, with Python 2.7.9+, 3.4+ are supported."

platform_system = platform.system()
if platform_system in ('Darwin', 'Linux', 'Windows', 'Microsoft', 'AIX'):
    if version_info[0] == 3:
        if version_info < (3, 4, 0):
            raise Exception(ERROR_STRING)
    elif version_info[0] == 2:
        if version_info[1] != 7:
            raise Exception(ERROR_STRING)
    else:
        raise Exception(ERROR_STRING)
else:
    warnings.warn("docplex is not officially supported on this platform. Use it at your own risk.", RuntimeWarning)

is_64bits = sys.maxsize > 2**32
if is_64bits is False:
    warnings.warn("docplex is not officially supported on 32 bits. Use it at your own risk.", RuntimeWarning)

from docplex.version import docplex_version_major, docplex_version_minor, docplex_version_micro
__version_info__ = (docplex_version_major, docplex_version_minor, docplex_version_micro)



# --------------------------------------------------------------------------
# Source file provided under Apache License, Version 2.0, January 2004,
# http://www.apache.org/licenses/
# (c) Copyright IBM Corp. 2015, 2019
# --------------------------------------------------------------------------
'''This package contains classes to monitor the progress of a MIP solve.

This abstract class defines protocol methods, which are called from within CPLEX's MIP search algorithm.

Progress listeners are based on CPLEX's MIPInfoCallback, and work only with a
local installation of CPLEX, not on Cplexcloud.

At each node, all progress listeners attached to the model receive progress data
through a `notify_progress` method, with an instance of :class:`~ProgressData`.
This named tuple class contains information about solve time, objective value,
best bound, gap and nodes, but no variable values.

The base class for progress listeners is :class:`~ProgressListener`.
DOcplex provides concrete sub-classes to handle basic cases, but
you may also create your own sub-class of :class:`~ProgressListener` to suit your needs.
Each instance of :class:`~ProgressListener` must be attached to a model to receive events during
a MIP solve., using `Model.add_progress_listener`.

An example of such a progress listener is :class:`TextProgressListener`,
which prints a message to stdout in a format similar to CPLEX log, each time it receives a progress data.

If you need to get the values of variables in an intermediate solution,
you should sub-class from class :class:`~SolutionListener`.
In this case, the :func:`~ProgressListener.notify_solution` method will be called
each time CPLEX finds a valid intermediate solution, with an instance of :class:`docplex.mp.solution.SolveSolution`.

Listeners are called from within CPLEX code, with its own frequency. It may occur that listeners are
called with no real progress in either the objective or the best bound. To help you chose what kind
of calls you are really interested in, DOcplex uses the enumerated class :class:`~ProgressClock'.
The baseline clock is the CPLEX clock, but you can refine this clock: for example,
`ProgressClock.Objective` filters all calls but those where the objective has improved.

To summarize, listeners are called from within CPLEX's MIP search algorithm, when the clock
you have chosen accepts the call. All listeners are created with a clock argument that controls which
calls they accept.

When a listener accepts a call, its method :func:`~ProgressListener.notify_progress` is called with the current progress data.
See :class:`~TextProgressListener` as an example of a simple progress listener.

In addition, if the listener derives from :class:`~SolutionListener`, its method :func:`~ProgressListener.notify_solution` is also called
with the intermediate solution, passed as an instance of :class:`docplex.mp.solution.SolveSolution`.
See :class:`~SolutionRecorder` as an example of solution listener

'''
from six import iteritems
from enum import Enum
from collections import namedtuple


from docplex.mp.utils import is_string, is_number
from docplex.mp.error_handler import docplex_fatal

_TProgressData_ = namedtuple('_TProgressData',
                             ['id', 'has_incumbent',
                              'current_objective', 'best_bound', 'mip_gap',
                              'current_nb_iterations', 'current_nb_nodes', 'remaining_nb_nodes',
                              'time', 'det_time'])


# noinspection PyUnresolvedReferences
class ProgressData(_TProgressData_):
    """ A named tuple class to hold progress data, as reeived from CPLEX.

    Attributes:
        has_incumbent: a boolean, indicating whether an incumbent solution is available (or not),
            at the moment the listener is called.
        current_objective: contains the current objective, if an incumbent is available, else None.
        best_bound: The current best bound as reported by the solver.
        mip_gap: the gap between the best integer objective and the objective of the best node remaining;
            available if an incumbent is available.
        current_nb_nodes: the current number of nodes.
        current_nb_iterations: the current number of iterations.
        remaining_nb_nodes: the remaining number of nodes.
        time: the elapsed time since solve started.
        det_time: the deterministic time since solve started.
    """

    def __str__(self):  # pragma: no cover
        fmt = 'ProgressData({0}, {1}, obj={2}, bbound={3}, #nodes={4})'. \
            format(self.id, self.has_incumbent, self.current_objective, self.best_bound, self.current_nb_nodes)
        return fmt


class ProgressClock(Enum):
    """
    This enumerated class controls the type of events a listener listens to.

    The possible values are (in order of decreasing call frequency).

        - All: the listener listens to all calls from CPLEX.
        - BestBound: listen to changes in best bound, not necessarily with a solution present.
        - Solutions: listen to all intermediate solutions, not necessarily impriving.
            Nothing prevents being called several times with an identical solution.

        - Gap: listen to intermediate solutions, where either objective or best bound has improved.
        - Objective: listen to intermediate solutions, where the solution objective has improved.

    To determine whether objective or best bound have improved (or not), see the numerical parameters
    `absdiff` and `reldiff` in `:class:ProgressListener` and all its descendant classes.

    *New in version 2.10*
    """
    #
    # An enumeration of filtering levels for listeners.
    #
    All = 0
    Solutions = 1
    BestBound = 2  # best bound clock is indepedent from solution clock
    Objective = 5  # b101
    Gap = 7  # the gap clock changes when there is a solution, and either bound or objective has changed

    @property
    def listens_to_solution(self):
        # returns true if the enum listens to solutions
        return 1 == self.value & 1

    @classmethod
    def parse(cls, arg):
        if isinstance(arg, ProgressClock):
            return arg
        else:
            # int value
            for fl in cls:
                if arg == fl.value:
                    return fl
                elif is_string(arg) and arg.lower() == fl.name.lower():
                    return fl
            else:
                # pragma: no cover
                raise ValueError('Expecting filter level, {0!r} was passed'.format(arg))


default_absdiff = 1e-1
default_reldiff = 1e-2


class ProgressListener(object):
    '''  The base class for progress listeners.
    '''

    def __init__(self, clock_arg=ProgressClock.All, absdiff=None, reldiff=None):
        self._cb = None
        clock = ProgressClock.parse(clock_arg)
        self._clock = clock
        self._current_progress_data = None
        self._absdiff = absdiff if absdiff is not None else default_absdiff
        self._reldiff = reldiff if reldiff is not None else default_reldiff
        self._filter = make_clock_filter(clock, absdiff, reldiff)

    def _get_model(self):
        ccb = self._cb
        return ccb._model if ccb else None

    @property
    def clock(self):
        """ Returns the clock of the listener.

        :return:
            an instance of :class:`~ProgressClock`

        """
        return self._clock

    @property
    def abs_diff(self):
        return self._absdiff

    @property
    def relative_diff(self):
        return self._reldiff

    @property
    def current_progress_data(self):
        """ This property return the current progress data, if any.

        Returns the latest progress data, if at least one has been received.

        :return: an instance of :class:~ProgressData` or None.
        """
        return self._current_progress_data

    def _set_current_progress_data(self, pdata):
        # INTERNAL
        self._current_progress_data = pdata

    def accept(self, pdata):
        return self._filter.accept(pdata)

    def _disconnect(self):
        # INTERNAL
        self._cb = None

    def _connect_cb(self, cb):
        # INTERNAL
        self._cb = cb

    def requires_solution(self):
        return False

    def notify_solution(self, s):
        """ Redefine this method to handle an intermediate solution from the callback.

        Args:
            s: solution
        """
        pass  # pragma: no cover

    def notify_start(self):
        """ This methods  is called on all attached listeners at the beginning of a solve().

        When writing a custom listener, this method is used to restore
        internal data which can be modified during solve.
        Always call the superclass `notify_start` before adding specific code in a sub-class.
        """
        self._current_progress_data = None
        self._filter.reset()

    def notify_jobid(self, jobid):
        pass  # pragma: no cover

    def notify_end(self, status, objective):
        """The method called when solve is finished on a model.

        The status is the solve status from the
        solve() method
        """
        pass  # pragma: no cover

    def notify_progress(self, progress_data):
        """ This method is called from within the solve with a ProgressData instance.

        :param progress_data: an instance of :class:`ProgressData` containing data about the
            current point in the search tree.

        """
        pass  # pragma: no cover

    def abort(self):
        ''' Aborts the CPLEX search.

        This method tells CPLEX to stop the MIP search.
        You may use this method in a custom progress listener to stop the search based on your own
        criteria (for example, when improvements in gap is consistently below a minimum threshold).

        Note:
            This method should be called from within a :func:`notify_progress` method call, otherwise it will have
            no effect.

        '''
        ccb = self._cb
        if ccb is not None:
            ccb.abort()
        else:
            print('!!! callback is not connected - abort() ignored')


# noinspection PyUnusedLocal,PyMethodMayBeStatic
class FilterAcceptAll(object):

    def accept(self, pdata):
        return True

    def reset(self):
        pass


# noinspection PyMethodMayBeStatic
class FilterAcceptAllSolutions(object):

    def accept(self, pdata):
        return pdata.has_incumbent

    def reset(self):
        pass


class Watcher(object):

    def __init__(self, name, absdiff, reldiff, update_fn):
        assert absdiff >= 0
        assert reldiff >= 0

        self.name = name
        self._watched = None
        self._old = None
        self._absdiff = absdiff
        self._reldiff = reldiff
        self._update_fn = update_fn

    def reset(self):
        self._watched = None
        self._old = None

    def accept(self, progress_data):
        accepted = False

        old = self._watched
        new_watched = self._update_fn(progress_data)
        # if new_watched is None, then it is not available (e.g. objective)
        if new_watched is not None:
            if old is None:
                accepted = True
            else:
                # assert is a number
                delta = abs(new_watched - old)
                reldiff = self._reldiff
                absdiff = self._absdiff
                if absdiff > 0 and abs(new_watched - old) >= absdiff:
                    accepted = True
                elif reldiff and delta / (1 + abs(old)) >= reldiff:
                    accepted = True
        return accepted

    def sync(self, pdata):
        cur = self._watched
        self._watched = self._update_fn(pdata)
        self._old = cur

    def __str__(self):  # pragma: no cover
        ws = '--' if self._watched is None else self._watched
        return 'W_{0}[{1}]'.format(self.name, ws)

def clock_filter_accept_stop_here():
    pass

class ClockFilter(object):

    def __init__(self, level, obj_absdiff, bbound_absdiff, obj_reldiff=0, bbound_reldiff=0, node_delta=-1):
        watchers = []
        if obj_absdiff > 0 or obj_reldiff > 0:
            def update_obj(pdata):
                return pdata.current_objective if pdata.has_incumbent else None

            obj_watcher = Watcher(name='obj', absdiff=obj_absdiff, reldiff=obj_reldiff, update_fn=update_obj)
            watchers.append(obj_watcher)
        if bbound_absdiff > 0 or bbound_reldiff > 0:
            def update_bbound(pdata):
                return pdata.best_bound

            watchers.append(Watcher(name='gap_bound', absdiff=bbound_absdiff, reldiff=bbound_reldiff,
                                    update_fn=update_bbound))
        if node_delta > 0:
            used_node_delta = max(node_delta, 1)

            def update_nodes(pdata):
                return pdata.current_nb_nodes

            watchers.append(Watcher(name='nodes', absdiff=used_node_delta, reldiff=0, update_fn=update_nodes))
        self._watchers = watchers
        self._clock = level
        self._listens_to_solution = level.listens_to_solution

    def accept(self, progress_data):
        if not progress_data.has_incumbent and self._listens_to_solution:
            return False
        # the filter accepts data as soon as one of its cells accepts it.
        poke = self.peek(progress_data)
        if poke is None:
            return False
        else:
            clock_filter_accept_stop_here()
            # print("- accepting event #{0}, reason: {1}".format(progress_data.id, poke.name))
            for w in self._watchers:
                w.sync(progress_data)
            return True

    def peek(self, progress_data):
        for w in self._watchers:
            if w.accept(progress_data):
                return w
        else:
            return None

    def reset(self):
        for w in self._watchers:
            w.reset()


# noinspection PyArgumentEqualDefault
def make_clock_filter(level, absdiff, reldiff, nodediff=0):
    absdiff_ = 1e-1 if absdiff is None else absdiff
    reldiff_ = 1e-2 if reldiff is None else reldiff
    if level == ProgressClock.All:
        return FilterAcceptAll()
    elif level == ProgressClock.Solutions:
        return FilterAcceptAllSolutions()
    elif level == ProgressClock.Gap:
        return ClockFilter(level, obj_absdiff=absdiff_, obj_reldiff=reldiff_,
                           bbound_absdiff=absdiff_, bbound_reldiff=reldiff_, node_delta=nodediff)
    elif level == ProgressClock.Objective:
        return ClockFilter(level, obj_absdiff=absdiff_, obj_reldiff=reldiff_,
                           bbound_absdiff=0, bbound_reldiff=0, node_delta=nodediff)
    elif level == ProgressClock.BestBound:
        return ClockFilter(level, obj_absdiff=0, obj_reldiff=0,
                           bbound_absdiff=absdiff_, bbound_reldiff=reldiff_)

    else:
        # pragma: no cover
        raise ValueError('unexpected level: {0!r}'.format(level))


class TextProgressListener(ProgressListener):
    """ A simple implementation of Progress Listener, which prints messages to stdout,
        in the manner of the CPLEX log.

    :param clock: an enumerated value of type `:class:ProgressClock` which defines the frequency of the listener.
    :param absdiff: a float value which controls the minimum absolute change for objective or best bound values
    :param reldiff: a float value which controls the minimum absolute change for objective or best bound values.

    Note:
        The default behavior is to listen to an improvement in either the objective or best bound,
        each being improved by either an absolute value change of at least `absdiff`,
        or a relative change of at least `reldiff`.

    """

    def __init__(self, clock=ProgressClock.Gap, gap_fmt=None, obj_fmt=None,
                 absdiff=None, reldiff=None):
        ProgressListener.__init__(self, clock, absdiff, reldiff)
        self._gap_fmt = gap_fmt or "{:.2%}"
        self._obj_fmt = obj_fmt or "{:.4f}"
        self._count = 0

    def notify_start(self):
        super(TextProgressListener, self).notify_start()
        self._count = 0

    def notify_progress(self, progress_data):
        self._count += 1
        pdata_has_incumbent = progress_data.has_incumbent
        incumbent_symbol = '+' if pdata_has_incumbent else ' '
        # if pdata_has_incumbent:
        #     self._incumbent_count += 1
        current_obj = progress_data.current_objective
        if pdata_has_incumbent:
            objs = self._obj_fmt.format(current_obj)
        else:
            objs = "N/A"  # pragma: no cover
        best_bound = progress_data.best_bound
        nb_nodes = progress_data.current_nb_nodes
        remaining_nodes = progress_data.remaining_nb_nodes
        if pdata_has_incumbent:
            gap = self._gap_fmt.format(progress_data.mip_gap)
        else:
            gap = "N/A"  # pragma: no cover
        raw_time = progress_data.time
        rounded_time = round(raw_time, 1)

        print("{0:>3}{7}: Node={4} Left={5} Best Integer={1}, Best Bound={2:.4f}, gap={3}, ItCnt={8} [{6}s]"
              .format(self._count, objs, best_bound, gap, nb_nodes, remaining_nodes, rounded_time,
                      incumbent_symbol, progress_data.current_nb_iterations))


class ProgressDataRecorder(ProgressListener):
    """ A specialized class of ProgressListener, which collects all ProgressData it receives.

    """

    def __init__(self, clock=ProgressClock.Gap, absdiff=None, reldiff=None):
        super(ProgressDataRecorder, self).__init__(clock, absdiff, reldiff)
        self._recorded = []

    def notify_start(self):
        super(ProgressDataRecorder, self).notify_start()
        # clear recorded data
        self._recorded = []

    def notify_progress(self, progress_data):
        self._recorded.append(progress_data)

    @property
    def number_of_records(self):
        return len(self._recorded)

    @property
    def iter_recorded(self):
        """ Returns an iterator on stored progress data

        :return: an iterator.
        """
        return iter(self._recorded)

    @property
    def recorded(self):
        """ Returns a copy of the recorded data.

        :return:
        """
        return self._recorded[:]


class SolutionListener(ProgressListener):
    """ The base class for listeners that work on intermediate solutions.

    To define a custom behavior for a subclass of this class, you need to redefine `notify_solution`.
    The current progress data is available from the `current_progress_data` property.
    """

    # noinspection PyMethodMayBeStatic
    def check_solution_clock(self):
        if not self.clock.listens_to_solution:
            docplex_fatal('Solution listener requires a solution clock among (Solutions,Objective|Gap), {0} was passed',
                          self.clock)

    def __init__(self, clock=ProgressClock.Solutions, absdiff=None, reldiff=None):
        super(SolutionListener, self).__init__(clock, absdiff, reldiff)
        self.check_solution_clock()

    def requires_solution(self):
        return True

    def notify_solution(self, sol):
        """ Generic method to be redefined by custom listeners to
        handle intermediate solutions. This method is called by the
        CPLEX search with an intermediate solution.

        :param sol: an instance of :class:`docplex.mp.solution.SolveSolution`

        Note: as this method is called at each node of the MIP search, it may happen
        that several calls are made with an identical solution, that is, different object
        instances, but sharing the same variable values.
        """
        pass  # pragma: no cover

    def notify_start(self):
        super(SolutionListener, self).notify_start()

    def accept(self, pdata):
        return pdata.has_incumbent and super(SolutionListener, self).accept(pdata)


class SolutionRecorder(SolutionListener):
    """ A specialized implementation of :class:`SolutionListener`,
    which stores --all-- intermediate solutions.

    As  the listener might be called at different times with identical incumbent
    values, thus the list of solutions list might well contain identical solutions.
    """

    def __init__(self, clock=ProgressClock.Gap, absdiff=None, reldiff=None):
        super(SolutionRecorder, self).__init__(clock, absdiff, reldiff)
        self._solutions = []

    def notify_start(self):
        """ Redefinition of the generic notify_start() method to clear all data modified by solve().
        In this case, clears the list of solutions.
        """
        super(SolutionListener, self).notify_start()
        self._solutions = []

    def notify_solution(self, sol):
        """ Redefintion of the generic `notify_solution` method, called by CPLEX.
         For this class, appends the intermediate solution to the list of stored solutions.

        """
        self._solutions.append(sol)

    def iter_solutions(self):
        """ Returns an iterator on the stored solutions"""
        return iter(self._solutions)

    @property
    def number_of_solutions(self):
        """ Returns the number of stored solutions. """
        return len(self._solutions)

    @property
    def current_solution(self):
        # redefinition of generic method `notify_solution`, called by CPLEX.
        sols = self._solutions
        return sols[-1] if sols else None


class FunctionalSolutionListener(SolutionListener):
    """ A subclass of SolutionListener, which calls a function at each intermediate solution.

        No exception is caught.
    """

    def __init__(self, solution_fn, clock=ProgressClock.Gap, absdiff=None, reldiff=None):
        SolutionListener.__init__(self, clock, absdiff, reldiff)
        self._sol_fn = solution_fn

    def notify_solution(self, sol):
        self._sol_fn(sol)


class KpiListener(SolutionListener):
    """ A refinement of SolutionListener, which computes KPIs at each intermediate solution.

    Calls the `publish` method with a dicitonary of KPIs. Defaul tis to do nothing.

    This listener listens to the `Gap` clock.

    """

    objective_kpi_name = '_current_objective'
    time_kpi_name = '_current_time'

    def __init__(self, model, clock=ProgressClock.Gap, absdiff=None, reldiff=None):
        super(KpiListener, self).__init__(clock, absdiff, reldiff)
        self.model = model

    def publish(self, kpi_dict):
        """ This method is called at each improving solution, with a dictionay of name, values.

        :param kpi_dict: a dicitonary of names and KPi values, computed on the intermediate solution.

        """
        pass

    def notify_solution(self, sol):
        pdata = self.current_progress_data

        # 1. build a dict from formatted names to kpi values.
        kpis_as_dict = {kp.name: kp.compute(sol) for kp in self.model.iter_kpis()}
        # 2. add predefined keys for obj, time.
        kpis_as_dict[self.objective_kpi_name] = sol.objective_value
        kpis_as_dict[self.time_kpi_name] = pdata.time
        self.publish(kpis_as_dict)

class KpiPrinter(KpiListener):

    def __init__(self, model, clock=ProgressClock.Gap, absdiff=None, reldiff=None,
                 kpi_format='* ItCnt={3:d}  KPI: {1:<{0}} = '):
        super(KpiPrinter, self).__init__(model, clock, absdiff, reldiff)
        self.kpi_format = kpi_format

    def publish(self, kpi_dict):
        try:
            max_kpi_name_len = max(len(kn) for kn in kpi_dict)  # max() raises ValueError on empty
        except ValueError:
            max_kpi_name_len = 0
        kpi_num_format = self.kpi_format + '{2:.3f}'
        kpi_str_format = self.kpi_format + '{2!s}'
        print('-' * (max_kpi_name_len + 15))
        itcnt = self.current_progress_data.current_nb_iterations
        for kn, kv in iteritems(kpi_dict):
            if is_number(kv):
                k_format = kpi_num_format
            else:
                k_format = kpi_str_format
            kps = k_format.format(max_kpi_name_len, kn, kv, itcnt)
            print(kps)

